import itertools
import json
import py_compile
import shutil
import sys
import zipfile
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any

from packaging.tags import Tag
from packaging.utils import parse_wheel_filename

from pyodide_build.common import _get_sha256_checksum
from pyodide_build.logger import logger, set_log_level


def _specialize_convert_tags(tags: set[Tag] | frozenset[Tag], wheel_name: str) -> Tag:
    """Convert a sequence of wheel tags to a single tag corresponding
    to the current interpreter and compatible with the py -> pyc compilation.

    Having more than one output tag is not supported.

    Examples
    --------
    >>> from packaging.tags import parse_tag
    >>> tags = parse_tag("py2.py3-none-any")
    >>> import re
    >>> re.sub("cp[0-9]*", "cpxxx", str(_specialize_convert_tags(set(tags), "")))
    'cpxxx-none-any'
    >>> tags = parse_tag("cp311-cp311-emscripten_3_1_24_wasm32")
    >>> re.sub("cp[0-9]*", "cpxxx", str(_specialize_convert_tags(set(tags), "")))
    'cpxxx-cpxxx-emscripten_3_1_24_wasm32'
    >>> tags = parse_tag("py310.py311-any-none")
    >>> re.sub("cp[0-9]*", "cpxxx", str(_specialize_convert_tags(set(tags), "")))
    'cpxxx-any-none'
    >>> tags = parse_tag("py36-abi3-none")
    >>> re.sub("cp[0-9]*", "cpxxx", str(_specialize_convert_tags(set(tags), "")))
    'cpxxx-abi3-none'
    """
    if len(tags) == 0:
        raise ValueError("Failed to parse tags from the wheel file name: {wheel_name}!")

    output_tags = set()
    interpreter = "cp" + "".join(str(el) for el in sys.version_info[:2])
    for tag in tags:
        output_tags.add(
            Tag(interpreter=interpreter, abi=tag.abi, platform=tag.platform)
        )

    if len(output_tags) > 1:
        # See https://github.com/pypa/packaging/issues/616
        raise NotImplementedError(
            "Found more than one output tag after py-compilation: "
            f"{[str(tag) for tag in output_tags]} in {wheel_name}"
        )

    return list(output_tags)[0]


def _py_compile_wheel_name(wheel_name: str) -> str:
    """Return the name of the py-compiled wheel

    See https://peps.python.org/pep-0427/ for more information.

    Examples
    --------
    >>> import re
    >>> re.sub("cp[0-9]*", "cpxxx", _py_compile_wheel_name('micropip-0.1.0-py3-none-any.whl'))
    'micropip-0.1.0-cpxxx-none-any.whl'
    >>> re.sub("cp[0-9]*", "cpxxx", _py_compile_wheel_name("numpy-1.22.4-cp311-cp311-emscripten_3_1_24_wasm32.whl"))
    'numpy-1.22.4-cpxxx-cpxxx-emscripten_3_1_24_wasm32.whl'
    >>> # names with '_' are preserved (instead of using '-')
    >>> re.sub("cp[0-9]*", "cpxxx", _py_compile_wheel_name("a_b-0.0.0-cp311-cp311-emscripten_3_1_24_wasm32.whl"))
    'a_b-0.0.0-cpxxx-cpxxx-emscripten_3_1_24_wasm32.whl'
    >>> # if there are multiple tags (e.g. py2 & py3), we only keep the relevant one
    >>> re.sub("cp[0-9]*", "cpxxx", _py_compile_wheel_name('attrs-21.4.0-py2.py3-none-any.whl'))
    'attrs-21.4.0-cpxxx-none-any.whl'


    # >>> msg = "Processing more than one tag is not implemented"
    # >>> with pytest.rases(NotImplementedError, match=msg):
    # ...     _py_compile_wheel_name("numpy-1.23.4-cp311-cp311-manylinux_2_17_aarch64.manylinux2014_aarch64.whl")
    """
    name, version, build, tags = parse_wheel_filename(wheel_name)
    if build:
        # TODO: not sure what to do here, but we never have such files in Pyodide
        # Opened https://github.com/pypa/packaging/issues/616 about it.
        raise NotImplementedError(f"build tag {build} not implemented")
    output_name = f"{name.replace('-', '_')}-{version}-"
    output_name += str(_specialize_convert_tags(tags, wheel_name=wheel_name))
    return output_name + ".whl"


def _compile(
    input_path: Path,
    output_path: Path,
    keep: bool = True,
    verbose: bool = True,
    compression_level: int = 6,
) -> None:
    """Compile all .py files in the zip archive to .pyc files.

    Parameters
    ----------
    input_path
        Path to the input archive.
    output_path
        Path to the output archive.
    compression_level
        Level of zip compression to apply. 0 means no compression. If a strictly
        positive integer is provided, ZIP_DEFLATED option is used.
    """
    output_name = output_path.name

    with set_log_level(logger, verbose):
        logger.debug("Running py-compile on %s to %s", input_path, output_path)

        if compression_level > 0:
            compression = zipfile.ZIP_DEFLATED
        else:
            compression = zipfile.ZIP_STORED

        with (
            zipfile.ZipFile(input_path) as fh_zip_in,
            TemporaryDirectory() as temp_dir_str,
        ):
            temp_dir = Path(temp_dir_str)
            output_path_tmp = temp_dir / output_name
            with zipfile.ZipFile(
                output_path_tmp,
                mode="w",
                compression=compression,
                compresslevel=compression_level,
            ) as fh_zip_out:
                for name in fh_zip_in.namelist():
                    if name.endswith(".pyc"):
                        # We are going to re-compile all .pyc files
                        continue

                    stream = fh_zip_in.read(name)
                    if not name.endswith(".py"):
                        # Write file without changes
                        fh_zip_out.writestr(name, stream)
                        continue

                    # Otherwise write file to disk and run py_compile
                    # Unfortunately py_compile doesn't support bytes input/output, it has to be real files
                    tmp_path_py = temp_dir / name.replace("/", "_")
                    tmp_path_py.write_bytes(stream)

                    tmp_path_pyc = temp_dir / (tmp_path_py.name + "c")
                    py_compile.compile(
                        str(tmp_path_py),
                        cfile=str(tmp_path_pyc),
                        dfile=name,
                        doraise=True,
                    )

                    fh_zip_out.writestr(name + "c", tmp_path_pyc.read_bytes())
            if output_path == input_path:
                if keep:
                    logger.debug("Adding .old suffix to avoid overwriting input file.")

                    backup_path = input_path.with_suffix(input_path.suffix + ".old")
                    input_path.rename(backup_path)
            elif not keep:
                # Remove input file
                input_path.unlink()

            shutil.copyfile(output_path_tmp, output_path)


def _py_compile_archive(
    input_path: Path,
    keep: bool = True,
    verbose: bool = True,
    compression_level: int = 6,
) -> Path | None:
    """Compile .py files to .pyc in a wheel or zip file.

    All non Python files are kept unchanged.

    Parameters
    ----------
    input_path
        input path to a .whl or .zip file
    keep
        if False, delete the input file. Otherwise, it will be either kept or
        renamed with a suffix .whl.old (if the input path == computed output
        path)
    verbose
        print logging information
    compression_level
        Level of zip compression to apply. 0 means no compression. If a strictly
        positive integer is provided, ZIP_DEFLATED option is used.

    Returns
    -------
    path
        path to processed archive with .pyc files. Or None if the file was not py-compiled
        (e.g. if it's a zip with no .py files)
    """
    if input_path.suffix not in [".whl", ".zip"]:
        raise ValueError(
            f"Error: only .whl or .zip files are supported, got {input_path.name}"
        )

    if not input_path.exists():
        raise FileNotFoundError(f"{input_path} does not exist!")

    name_out = _get_py_compiled_archive_name(input_path)
    if name_out is None:
        return None
    path_out = input_path.parent / name_out

    _compile(
        input_path,
        path_out,
        keep=keep,
        verbose=verbose,
        compression_level=compression_level,
    )

    return path_out


def _get_py_compiled_archive_name(path: Path) -> str | None:
    """Return the name of the py-compiled wheel or zip file

    Returns None if the file should not be py-compiled.

    Examples
    --------
    >>> import re
    >>> re.sub("cp[0-9]*", "cpxxx", _get_py_compiled_archive_name(Path("snowballstemmer-2.2.0-py2.py3-none-any.whl")))
    'snowballstemmer-2.2.0-cpxxx-none-any.whl'
    """

    if path.suffix == ".whl":
        try:
            output_name = _py_compile_wheel_name(path.name)
            return output_name
        except Exception as e:
            print(e)
            return None
    elif path.suffix == ".zip":
        # If it's a zip file with .py files, keep the same name
        with zipfile.ZipFile(path, "r") as zip_ref:
            if any(file.endswith(".py") for file in zip_ref.namelist()):
                return path.name
        return None
    else:
        return None


def _update_lockfile(
    input_dir: Path, lockfile: dict[str, Any], name_mapping: dict[str, str]
) -> dict[str, Any]:
    """Update pyodide-lock.json with the new names of the py-compiled wheels.

    Also update the checksums of the updated wheels
    """
    for row in lockfile["packages"].values():
        if row.get("file_name") in name_mapping:
            row["file_name"] = name_mapping[row["file_name"]]
            row["sha256"] = _get_sha256_checksum(input_dir / row["file_name"])
    return lockfile


def _py_compile_archive_dir(
    input_dir: Path,
    keep: bool = True,
    verbose: bool = True,
    compression_level: int = 6,
    excludes: list[str] | None = None,
) -> dict[str, str]:
    """Py-compile all wheels or zip files in a directory.

    All .py files in the wheels or zip files  are compiled to .pyc files. All
    non Python files are kept unchanged.
    For wheel the file names will be changed to include the Python version used
    for the compilation following the PEP 425 convention.


    Parameters
    ----------
    wheel_path
        input wheel path
    keep
        if False, delete the input file. Otherwise, it will be either kept or
        renamed with a suffix .whl.old (if the input path == computed output
        path)
    verbose
        print logging information
    compression_level
        Level of zip compression to apply. 0 means no compression. If a strictly
        positive integer is provided, ZIP_DEFLATED option is used.

    Returns
    -------
    name_mapping
        mapping between old and new file names
    """

    name_mapping = {}

    for file_path in itertools.chain(
        *[input_dir.glob(ext) for ext in ["*.zip", "*.whl"]]
    ):
        if excludes and any(file_path.name.startswith(exclude) for exclude in excludes):
            continue

        output_name = _get_py_compiled_archive_name(file_path)
        if output_name is None:
            continue

        _compile(
            file_path,
            file_path.parent / output_name,
            keep=keep,
            verbose=verbose,
            compression_level=compression_level,
        )
        name_mapping[file_path.name] = output_name

    lockfile_path = input_dir / "pyodide-lock.json"
    if name_mapping and lockfile_path.exists():
        if verbose:
            print(f"Updating {lockfile_path.name}")
        with open(lockfile_path) as fh:
            lockfile = json.load(fh)
        lockfile = _update_lockfile(input_dir, lockfile, name_mapping)
        with open(lockfile_path, "w") as fh:
            json.dump(lockfile, fh)
    return name_mapping



import os
import re
import subprocess
from pathlib import Path
from textwrap import dedent


def fix_f2c_input(f2c_input: Path) -> None:
    if f2c_input.name.endswith("_flapack-f2pywrappers.f"):
        content = f2c_input.read_text()
        content = content.replace("character cmach", "integer cmach")
        content = content.replace("character norm", "integer norm")
        f2c_input.write_text(content)
        return

    if f2c_input.name in [
        "_lapack_subroutine_wrappers.f",
        "_blas_subroutine_wrappers.f",
    ]:
        content = f2c_input.read_text()
        content = content.replace("character", "integer")
        content = content.replace(
            "ret = chla_transtype(", "call chla_transtype(ret, 1,"
        )
        f2c_input.write_text(content)


def fix_f2c_output(f2c_output: Path) -> None:
    """
    This function is called on the name of each C output file. It fixes up the C
    output in various ways to compensate for the lack of f2c support for Fortran
    90 and Fortran 95.
    """
    if f2c_output.name == "_lapack_subroutine_wrappers.c":
        content = f2c_output.read_text()
        content = content.replace("integer chla_transtype__", "void chla_transtype__")
        f2c_output.write_text(content)
        return

    if f2c_output.name.endswith("eupd.c"):
        content = f2c_output.read_text()
        content = re.sub(
            r"ftnlen\s*(howmny_len|bmat_len),?", "", content, flags=re.MULTILINE
        )
        f2c_output.write_text(content)
        return

    if f2c_output.name.endswith("lansvd.c"):
        content = f2c_output.read_text()
        content += dedent(
            """
            #include <time.h>

            int second_(real *t) {
                *t = clock()/1000;
                return 0;
            }
            """
        )
        f2c_output.write_text(content)
        return


def scipy_fix_cfile(path: Path) -> None:
    """
    Replace void return types with int return types in various generated .c and
    .h files. We can't achieve this with a simple patch because these files are
    not in the sdist, they are generated as part of the build.
    """
    text = path.read_text()
    text = text.replace("extern void F_WRAPPEDFUNC", "extern int F_WRAPPEDFUNC")
    text = text.replace("extern void F_FUNC", "extern int F_FUNC")
    text = text.replace("void (*f2py_func)", "int (*f2py_func)")
    text = text.replace("static void cb_", "static int cb_")
    text = text.replace("typedef void(*cb_", "typedef int(*cb_")
    text = text.replace("void(*)", "int(*)")
    text = text.replace("static void f2py_setup_", "static int f2py_setup_")

    # Fix Cython-generated function pointer casts for __Pyx_IsSameCFunction
    # for _matfuncs_sqrtm for builds of SciPy 1.16.0.
    # TODO: I'm not sure if this is the best place to put this, but it works
    text = re.sub(
        r"(__Pyx_IsSameCFunction\([^,]+,\s*)\(int\(\*\)\(void\)\)",
        r"\1(void(*)(void))",
        text,
    )

    if path.name.endswith("_flapackmodule.c"):
        text = text.replace(",size_t", "")
        text = re.sub(r",slen\([a-z]*\)\)", ")", text)

    path.write_text(text)

    for lib in ["lapack", "blas"]:
        if path.name.endswith(f"cython_{lib}.c"):
            header_name = f"_{lib}_subroutines.h"
            header_dir = path.parent
            header_path = find_header(header_dir, header_name)

            header_text = header_path.read_text()
            header_text = header_text.replace("void F_FUNC", "int F_FUNC")
            header_path.write_text(header_text)


def find_header(source_dir: Path, header_name: str) -> Path:
    """
    Find the header file that corresponds to a source file.
    """
    while not (header_path := source_dir / header_name).exists():
        # meson copies the source files into a subdirectory of the build
        source_dir = source_dir.parent
        if source_dir == Path("/"):
            raise RuntimeError(f"Could not find header file {header_name}")

    return header_path


def scipy_fixes(args: list[str]) -> None:
    for arg in args:
        if arg.endswith(".c"):
            scipy_fix_cfile(Path(arg))


def replay_f2c(args: list[str], dryrun: bool = False) -> list[str] | None:
    """Apply f2c to compilation arguments

    Parameters
    ----------
    args
       input compiler arguments
    dryrun
       if False run f2c on detected fortran files

    Returns
    -------
    new_args
       output compiler arguments


    Examples
    --------

    >>> replay_f2c(['gfortran', 'test.f'], dryrun=True)
    ['gcc', 'test.c']
    """
    f2c_path = os.environ.get("F2C_PATH", "f2c")

    new_args = ["gcc"]
    found_source = False
    for arg in args[1:]:
        if not arg.endswith((".f", ".F")):
            new_args.append(arg)
            continue
        found_source = True
        filepath = Path(arg).resolve()
        new_args.append(arg[:-2] + ".c")
        if dryrun:
            continue
        fix_f2c_input(Path(arg))
        if arg.endswith(".F"):
            # .F files apparently expect to be run through the C
            # preprocessor (they have #ifdef's in them)
            # Use gfortran frontend, as gcc frontend might not be
            # present on osx
            # The file-system might be not case-sensitive,
            # so take care to handle this by renaming.
            # For preprocessing and further operation the
            # expected file-name and extension needs to be preserved.
            subprocess.check_call(
                [
                    "gfortran",
                    "-E",
                    "-C",
                    "-P",
                    filepath,
                    "-o",
                    filepath.with_suffix(".f77"),
                ]
            )
            filepath = filepath.with_suffix(".f77")
        # -R flag is important, it means that Fortran functions that
        # return real e.g. sdot will be transformed into C functions
        # that return float. For historic reasons, by default f2c
        # transform them into functions that return a double. Using -R
        # allows to match what OpenBLAS has done when they f2ced their
        # Fortran files, see
        # https://github.com/xianyi/OpenBLAS/pull/3539#issuecomment-1493897254
        # for more details
        with (
            open(filepath) as input_pipe,
            open(filepath.with_suffix(".c"), "w") as output_pipe,
        ):
            subprocess.check_call(
                [f2c_path, "-R"],
                stdin=input_pipe,
                stdout=output_pipe,
                cwd=filepath.parent,
            )
        fix_f2c_output(Path(arg[:-2] + ".c"))

    new_args_str = " ".join(args)
    if ".so" in new_args_str and "libgfortran.so" not in new_args_str:
        found_source = True

    if not found_source:
        return None
    return new_args



"""
Build all of the packages in a given directory.
"""

import dataclasses
import datetime
import shutil
import subprocess
import sys
from collections import defaultdict
from collections.abc import Iterable, Iterator
from contextlib import contextmanager
from functools import total_ordering
from graphlib import TopologicalSorter
from pathlib import Path
from queue import PriorityQueue, Queue
from threading import Lock, Thread
from time import perf_counter, sleep
from typing import Any

from packaging.utils import canonicalize_name
from pyodide_lock import PyodideLockSpec
from pyodide_lock.spec import PackageSpec as PackageLockSpec
from pyodide_lock.utils import update_package_sha256
from rich.live import Live
from rich.progress import BarColumn, Progress, TimeElapsedColumn
from rich.spinner import Spinner
from rich.table import Table

from pyodide_build import build_env, uv_helper
from pyodide_build.build_env import BuildArgs
from pyodide_build.common import (
    download_and_unpack_archive,
    exit_with_stdio,
    extract_wheel_metadata_file,
    find_matching_wheel,
    find_missing_executables,
    repack_zip_archive,
)
from pyodide_build.logger import console_stdout, logger
from pyodide_build.recipe import loader, unvendor
from pyodide_build.recipe.builder import needs_rebuild
from pyodide_build.recipe.spec import MetaConfig, _BuildSpecTypes


class BuildError(Exception):
    def __init__(self, returncode: int, msg: str) -> None:
        self.returncode = returncode
        self.msg = msg
        super().__init__()


@total_ordering
@dataclasses.dataclass(eq=False, repr=False)
class BasePackage:
    pkgdir: Path
    name: str
    version: str
    disabled: bool
    meta: MetaConfig
    package_type: _BuildSpecTypes
    run_dependencies: list[str]
    host_dependencies: list[str]
    executables_required: list[str]
    dependencies: set[str]  # run + host dependencies
    unbuilt_host_dependencies: set[str]
    host_dependents: set[str]
    unvendor_tests: bool = False
    file_name: str | None = None
    install_dir: str = "site"
    _queue_idx: int | None = None

    def __init__(self, pkgdir: Path, config: MetaConfig):
        self.pkgdir = pkgdir
        self.meta = config.model_copy(deep=True)

        self.name = self.meta.package.name
        self.version = self.meta.package.version
        self.disabled = self.meta.package.disabled
        self.package_type = self.meta.build.package_type
        self.is_wheel = self.package_type in ["package", "cpython_module"]

        assert self.name == pkgdir.name, f"{self.name} != {pkgdir.name}"

        self.run_dependencies = self.meta.requirements.run
        self.host_dependencies = self.meta.requirements.host
        self.executables_required = self.meta.requirements.executable
        self.dependencies = set(self.run_dependencies + self.host_dependencies)
        self.unbuilt_host_dependencies = set(self.host_dependencies)
        self.host_dependents = set()

    def __hash__(self) -> int:
        return hash((self.name, self.version))

    @classmethod
    def from_recipe(
        cls,
        pkgdir: Path,
        config: MetaConfig,
    ) -> "BasePackage":
        match config.build.package_type:
            case "package" | "cpython_module":
                return PythonPackage(pkgdir, config)
            case "shared_library":
                return SharedLibrary(pkgdir, config)
            case "static_library":
                return StaticLibrary(pkgdir, config)
            case _:
                raise ValueError(f"Unknown package type: {config.build.package_type}")

    # We use this in the priority queue, which pops off the smallest element.
    # So we want the smallest element to have the largest number of dependents
    def __lt__(self, other: Any) -> bool:
        return len(self.host_dependents) > len(other.host_dependents)

    def __eq__(self, other: Any) -> bool:
        return len(self.host_dependents) == len(other.host_dependents)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.name})"

    def build_path(self, build_dir: Path) -> Path:
        return build_dir / self.name / "build"

    def needs_rebuild(self, build_dir: Path) -> bool:
        res = needs_rebuild(
            self.pkgdir,
            self.build_path(build_dir),
            self.meta.source,
            self.is_wheel,
            self.version,
        )
        return res

    def build(self, build_args: BuildArgs, build_dir: Path) -> None:
        run_prefix = (
            [uv_helper.find_uv_bin(), "run"] if uv_helper.should_use_uv() else []
        )
        p = subprocess.run(
            [
                *run_prefix,
                "pyodide",
                "build-recipes-no-deps",
                self.name,
                "--recipe-dir",
                str(self.pkgdir.parent),
                f"--cflags={build_args.cflags}",
                f"--cxxflags={build_args.cxxflags}",
                f"--ldflags={build_args.ldflags}",
                f"--target-install-dir={build_args.target_install_dir}",
                f"--host-install-dir={build_args.host_install_dir}",
                f"--build-dir={build_dir}",
                # Either this package has been updated and this doesn't
                # matter, or this package is dependent on a package that has
                # been updated and should be rebuilt even though its own
                # files haven't been updated.
                "--force-rebuild",
                # We already did the rust setup in buildall
                "--skip-rust-setup",
            ],
            check=False,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )

        if p.returncode != 0:
            msg = []
            msg.append(f"Error building {self.name}. Printing build logs.")
            logfile = self.pkgdir / "build.log"
            if logfile.is_file():
                msg.append(logfile.read_text(encoding="utf-8") + "\n")
            else:
                msg.append("ERROR: No build log found.")
            msg.append(f"ERROR: cancelling buildall due to error building {self.name}")
            raise BuildError(p.returncode, "\n".join(msg))

    def dist_artifact_path(self) -> Path | None:
        """
        Return the path to the distribution artifact of the package.
        """
        raise NotImplementedError()


@dataclasses.dataclass
class PythonPackage(BasePackage):
    def __init__(self, pkgdir: Path, config: MetaConfig) -> None:
        super().__init__(pkgdir, config)

        self.unvendor_tests = self.meta.build.unvendor_tests

    def dist_artifact_path(self) -> Path | None:
        dist_dir = self.pkgdir / "dist"
        wheel = find_matching_wheel(
            dist_dir.glob("*.whl"), build_env.pyodide_tags(), version=self.version
        )
        if not wheel:
            raise RuntimeError(f"Found no wheel while building {self.name}")
        return wheel


@dataclasses.dataclass
class SharedLibrary(BasePackage):
    install_dir: str = "dynlib"

    def __init__(self, pkgdir: Path, config: MetaConfig) -> None:
        super().__init__(pkgdir, config)

    def dist_artifact_path(self) -> Path | None:
        dist_dir = self.pkgdir / "dist"
        candidates = list(dist_dir.glob("*.zip"))

        if len(candidates) != 1:
            raise RuntimeError(
                f"Unexpected number of wheels/archives {len(candidates)} when building {self.name}"
            )

        return candidates[0]


@dataclasses.dataclass
class StaticLibrary(BasePackage):
    def __init__(self, pkgdir: Path, config: MetaConfig) -> None:
        super().__init__(pkgdir, config)

    def dist_artifact_path(self) -> Path | None:
        return None


class PackageStatus:
    def __init__(
        self, *, name: str, idx: int, thread: int, total_packages: int
    ) -> None:
        self.pkg_name = name
        self.prefix = f"[{idx}/{total_packages}] (thread {thread})"
        self.status = Spinner("dots", style="red", speed=0.2)
        self.table = Table.grid(padding=1)
        self.table.add_row(f"{self.prefix} building {self.pkg_name}", self.status)
        self.finished = False

    def finish(self, success: bool, elapsed_time: float) -> None:
        time = datetime.datetime.fromtimestamp(elapsed_time, tz=datetime.UTC)
        if time.minute == 0:
            minutes = ""
        else:
            minutes = f"{time.minute}m "
        timestr = f"{minutes}{time.second}s"

        status = "built" if success else "failed"
        done_message = f"{self.prefix} {status} {self.pkg_name} in {timestr}"

        self.finished = True

        if success:
            logger.success(done_message)
        else:
            logger.error(done_message)

    def __rich__(self):
        return self.table


class ReplProgressFormatter:
    def __init__(self, num_packages: int) -> None:
        self.progress = Progress(
            "[progress.description]{task.description}",
            BarColumn(),
            "{task.completed}/{task.total} [progress.percentage]{task.percentage:>3.0f}%",
            "Time elapsed:",
            TimeElapsedColumn(),
        )
        self.task = self.progress.add_task("Building packages...", total=num_packages)
        self.packages: list[PackageStatus] = []
        self.reset_grid()

    def reset_grid(self):
        """Empty out the rendered grids."""
        self.top_grid = Table.grid()

        for package in self.packages:
            self.top_grid.add_row(package)

        self.main_grid = Table.grid()
        self.main_grid.add_row(self.top_grid)
        self.main_grid.add_row(self.progress)

    def add_package(
        self, *, name: str, idx: int, thread: int, total_packages: int
    ) -> PackageStatus:
        status = PackageStatus(
            name=name, idx=idx, thread=thread, total_packages=total_packages
        )
        self.packages.append(status)
        self.reset_grid()
        return status

    def remove_package(self, pkg: PackageStatus) -> None:
        self.packages.remove(pkg)
        self.reset_grid()

    def update_progress_bar(self):
        """Step the progress bar by one (to show that a package finished)"""
        self.progress.update(self.task, advance=1)

    def __rich__(self):
        return self.main_grid


def _validate_package_map(pkg_map: dict[str, BasePackage]) -> bool:
    # Check if dependencies are valid
    for pkg_name, pkg in pkg_map.items():
        for runtime_dep_name in pkg.run_dependencies:
            runtime_dep = pkg_map[runtime_dep_name]
            if runtime_dep.package_type == "static_library":
                raise ValueError(
                    f"{pkg_name} has an invalid dependency: {runtime_dep_name}. Static libraries must be a host dependency."
                )

    # Check executables required to build packages are available
    missing_executables = defaultdict(list)
    for name, pkg in pkg_map.items():
        for exe in find_missing_executables(pkg.executables_required):
            missing_executables[exe].append(name)

    if missing_executables:
        error_msg = "The following executables are missing in the host system:\n"
        for executable, pkgs in missing_executables.items():
            error_msg += f"- {executable} (required by: {', '.join(pkgs)})\n"

        raise RuntimeError(error_msg)

    return True


def _parse_package_query(query: list[str] | str | None) -> tuple[set[str], set[str]]:
    """
    Parse a package query string into a list of requested packages and a list of
    disabled packages.

    Parameters
    ----------
    query
        A list of packages to build, this can be a comma separated string.

    Returns
    -------
    A tuple of two lists, the first list contains requested packages, the second
    list contains disabled packages.

    Examples
    --------
    >>> _parse_package_query(None)
    (set(), set())
    >>> requested, disabled = _parse_package_query("a,b,c")
    >>> requested == {'a', 'b', 'c'}, disabled == set()
    (True, True)
    >>> requested, disabled = _parse_package_query("a,b,!c")
    >>> requested == {'a', 'b'}, disabled == {'c'}
    (True, True)
    >>> requested, disabled = _parse_package_query(["a", "b", "!c"])
    >>> requested == {'a', 'b'}, disabled == {'c'}
    (True, True)
    """
    if not query:
        query = []

    if isinstance(query, str):
        query = [el.strip() for el in query.split(",")]

    requested = set()
    disabled = set()

    for name in query:
        if not name:  # empty string
            continue

        if name.startswith("!"):
            disabled.add(name[1:])
        else:
            requested.add(name)

    return requested, disabled


def generate_dependency_graph(
    packages_dir: Path,
    requested: set[str],
    disabled: set[str] | None = None,
) -> dict[str, BasePackage]:
    """This generates a dependency graph for given packages.

    A node in the graph is a BasePackage object defined above, which maintains
    a list of dependencies and also dependents. That is, each node stores both
    incoming and outgoing edges.

    The dependencies and dependents are stored via their name, and we have a
    lookup table pkg_map: Dict[str, BasePackage] to look up the corresponding
    BasePackage object. The function returns pkg_map, which contains all
    packages in the graph as its values.

    Parameters
    ----------
    packages_dir
        A directory that contains packages
    requested
        A set of packages to build
    disabled
        A set of packages to not build

    Returns
    -------
    A dictionary mapping package names to BasePackage objects
    """

    pkg: BasePackage
    pkgname: str
    pkg_map: dict[str, BasePackage] = {}

    if not disabled:
        disabled = set()

    # Create dependency graph.
    # On first pass add all dependencies regardless of whether
    # disabled since it might happen because of a transitive dependency
    graph = {}
    all_recipes = loader.load_all_recipes(packages_dir)
    no_numpy_dependents = "no-numpy-dependents" in requested
    requested.discard("no-numpy-dependents")
    packages = requested.copy()

    while packages:
        pkgname = packages.pop()

        if pkgname not in all_recipes:
            raise ValueError(
                f"No metadata file found for the following package: {pkgname}"
            )

        pkg = BasePackage.from_recipe(packages_dir / pkgname, all_recipes[pkgname])
        pkg_map[pkgname] = pkg
        graph[pkgname] = pkg.dependencies
        for dep in pkg.dependencies:
            if pkg_map.get(dep) is None:
                packages.add(dep)

    # Traverse in build order (dependencies first then dependents)
    # Mark a package as disabled if they've either been explicitly disabled
    # or if any of its transitive dependencies were marked disabled.
    for pkgname in TopologicalSorter(graph).static_order():
        pkg = pkg_map[pkgname]
        if pkgname in disabled:
            pkg.disabled = True
            continue
        if no_numpy_dependents and "numpy" in pkg.dependencies:
            pkg.disabled = True
            continue
        for dep in pkg.dependencies:
            if pkg_map[dep].disabled:
                pkg.disabled = True
                break

    # Now traverse in reverse build order (dependents first then their
    # dependencies).
    # Locate the subset of packages that are transitive dependencies of packages
    # that are requested and not disabled.
    requested_with_deps = requested.copy()
    disabled_packages = set()
    for pkgname in reversed(list(TopologicalSorter(graph).static_order())):
        pkg = pkg_map[pkgname]
        if pkg.disabled:
            requested_with_deps.discard(pkgname)
            disabled_packages.add(pkgname)
            continue

        if pkgname not in requested_with_deps:
            continue

        requested_with_deps.update(pkg.dependencies)
        for dep in pkg.host_dependencies:
            pkg_map[dep].host_dependents.add(pkg.name)

    pkg_map = {name: pkg_map[name] for name in requested_with_deps}

    _validate_package_map(pkg_map)

    if disabled_packages:
        logger.warning(
            "The following packages are disabled: %s", ", ".join(disabled_packages)
        )

    return pkg_map


def job_priority(pkg: BasePackage) -> int:
    if pkg.name == "numpy":
        return 0
    else:
        return 1


def format_name_list(l: list[str]) -> str:
    """
    >>> format_name_list(["regex"])
    'regex'
    >>> format_name_list(["regex", "parso"])
    'regex and parso'
    >>> format_name_list(["regex", "parso", "jedi"])
    'regex, parso, and jedi'
    """
    if len(l) == 1:
        return l[0]
    most = l[:-1]
    if len(most) > 1:
        most = [x + "," for x in most]
    return " ".join(most) + " and " + l[-1]


def mark_package_needs_build(
    pkg_map: dict[str, BasePackage], pkg: BasePackage, needs_build: set[str]
) -> None:
    """
    Helper for generate_needs_build_set. Modifies needs_build in place.
    Recursively add pkg and all of its dependencies to needs_build.
    """
    if pkg.name in needs_build:
        return
    needs_build.add(pkg.name)
    for dep in pkg.host_dependents:
        mark_package_needs_build(pkg_map, pkg_map[dep], needs_build)


def generate_needs_build_set(
    pkg_map: dict[str, BasePackage], build_dir: Path
) -> set[str]:
    """
    Generate the set of packages that need to be rebuilt.

    This consists of:
    1. packages whose source files have changed since they were last built
       according to needs_rebuild, and
    2. packages which depend on case 1 packages.
    """
    needs_build: set[str] = set()
    for pkg in pkg_map.values():
        # Otherwise, rebuild packages that have been updated and their dependents.
        if pkg.needs_rebuild(build_dir):
            mark_package_needs_build(pkg_map, pkg, needs_build)
    return needs_build


class _GraphBuilder:
    """A class to manage state for build_from_graph.

    build_from_graph has a bunch of moving parts: a threadpool, a Rich Live
    session to display progress information to the terminal, and a job graph to
    keep track of. This class keeps track of all this state.

    The only public class is `run`.
    """

    def __init__(
        self,
        pkg_map: dict[str, BasePackage],
        build_args: BuildArgs,
        build_dir: Path,
        needs_build: set[str],
    ):
        self.pkg_map: dict[str, BasePackage] = pkg_map
        self.build_args: BuildArgs = build_args
        self.build_dir: Path = build_dir
        self.needs_build: set[str] = needs_build
        self.build_queue: PriorityQueue[tuple[int, BasePackage]] = PriorityQueue()
        self.built_queue: Queue[tuple[BasePackage, BaseException | None]] = Queue()
        self.lock: Lock = Lock()
        self.building_rust_pkg: bool = False
        self.queue_idx: int = 1
        self.progress_formatter: ReplProgressFormatter = ReplProgressFormatter(
            len(self.needs_build)
        )

    @contextmanager
    def _queue_index(self, pkg: BasePackage) -> Iterator[int | None]:
        """
        yield the queue_index for the current job or None if the job is a Rust
        job and the rust lock is currently held.

        Set up as a context manager just for the rust packages.
        """
        is_rust_pkg = pkg.meta.is_rust_package()
        with self.lock:
            queue_idx = self.queue_idx
            if is_rust_pkg and self.building_rust_pkg:
                # Don't build multiple rust packages at the same time.
                # See: https://github.com/pyodide/pyodide/issues/3565
                # Note that if there are only rust packages left in the queue,
                # this will keep pushing and popping packages until the current rust package
                # is built. This is not ideal but presumably the overhead is negligible.
                self.build_queue.put((job_priority(pkg), pkg))
                yield None
                return
            if is_rust_pkg:
                self.building_rust_pkg = True
            self.queue_idx += 1
        try:
            yield queue_idx
        finally:
            if is_rust_pkg:
                self.building_rust_pkg = False

    @contextmanager
    def _pkg_status_display(self, n: int, pkg: BasePackage) -> Iterator[None]:
        """Control the status information for the package.

        Prints the "[{pkg-num}/{total_packages}] (thread n) building {package_name}"
        message and when done prints "succeeded/failed building package ... in ... seconds"
        plus updates the progress info in the console.
        """
        idx = pkg._queue_idx
        assert idx
        pkg_status = self.progress_formatter.add_package(
            name=pkg.name,
            idx=idx,
            thread=n,
            total_packages=len(self.needs_build),
        )
        t0 = perf_counter()
        success = True
        try:
            yield
        except BaseException:
            success = False
            raise
        finally:
            pkg_status.finish(success, perf_counter() - t0)
            self.progress_formatter.remove_package(pkg_status)

    def _build_one(self, n: int, pkg: BasePackage) -> BaseException | None:
        try:
            with self._pkg_status_display(n, pkg):
                pkg.build(self.build_args, self.build_dir)
        except BaseException as e:
            return e
        else:
            return None

    def _builder(self, n: int) -> None:
        """This is the logic that controls a thread in the thread pool."""
        while True:
            pkg = self.build_queue.get()[1]
            with self._queue_index(pkg) as idx:
                if idx is None:
                    # Rust package and we're already building one.
                    # Release the GIL so new packages get queued
                    sleep(0.01)
                    continue
                pkg._queue_idx = idx
                res = self._build_one(n, pkg)
                self.built_queue.put((pkg, res))
                if res:
                    # Build failed, quit the thread.
                    # Note that all other threads just keep going for a bit
                    # longer until we call sys.exit
                    return
                # Release the GIL so new packages get queued
                sleep(0.01)

    def run(self, n_jobs: int, already_built: set[str]) -> None:
        """Build the graph with n_jobs threads

        Prepare the queue by locating packages with no deps, set up the cli
        progress display, start up the threads, and manage build queue.
        """
        for pkg_name in self.needs_build:
            pkg = self.pkg_map[pkg_name]
            if len(pkg.unbuilt_host_dependencies) == 0:
                self.build_queue.put((job_priority(pkg), pkg))

        num_built = len(already_built)
        with Live(self.progress_formatter, console=console_stdout):
            for n in range(0, n_jobs):
                Thread(target=self._builder, args=(n + 1,), daemon=True).start()

            while num_built < len(self.pkg_map):
                [pkg, err] = self.built_queue.get()
                if err:
                    raise err

                num_built += 1

                self.progress_formatter.update_progress_bar()

                for _dependent in pkg.host_dependents:
                    dependent = self.pkg_map[_dependent]
                    dependent.unbuilt_host_dependencies.remove(pkg.name)
                    if len(dependent.unbuilt_host_dependencies) == 0:
                        self.build_queue.put((job_priority(dependent), dependent))


def _run(cmd, *args, check=False, **kwargs):
    result = subprocess.run(cmd, *args, **kwargs, check=check)
    if result.returncode != 0:
        logger.error("ERROR: command failed %s", " ".join(cmd))
        exit_with_stdio(result)
    return result


def _ensure_rust_toolchain():
    rust_toolchain = build_env.get_build_flag("RUST_TOOLCHAIN")
    _run(["rustup", "toolchain", "install", rust_toolchain])
    _run(["rustup", "default", rust_toolchain])

    url = build_env.get_build_flag("RUST_EMSCRIPTEN_TARGET_URL")
    if not url:
        # Install target with rustup target add
        _run(
            [
                "rustup",
                "target",
                "add",
                "wasm32-unknown-emscripten",
                "--toolchain",
                rust_toolchain,
            ]
        )
        return

    # Now we are going to delete the normal wasm32-unknown-emscripten sysroot
    # and replace it with our wasm-eh version.
    # We place the "install_token" to indicate that our custom sysroot has been
    # installed and which URL we got it from.
    result = _run(
        ["rustup", "which", "--toolchain", rust_toolchain, "rustc"],
        capture_output=True,
        text=True,
    )

    toolchain_root = Path(result.stdout).parents[1]
    rustlib = toolchain_root / "lib/rustlib"
    install_token = rustlib / "wasm32-unknown-emscripten_install-url.txt"
    if install_token.exists() and install_token.read_text() == url:
        return
    shutil.rmtree(rustlib / "wasm32-unknown-emscripten", ignore_errors=True)
    download_and_unpack_archive(
        url, rustlib, "wasm32-unknown-emscripten target", exists_ok=True
    )
    install_token.write_text(url)


def build_from_graph(
    pkg_map: dict[str, BasePackage],
    build_args: BuildArgs,
    build_dir: Path,
    n_jobs: int = 1,
    force_rebuild: bool = False,
) -> None:
    """
    This builds packages in pkg_map in parallel, building at most n_jobs
    packages at once.

    We have a priority queue of packages we are ready to build (build_queue),
    where a package is ready to build if all its dependencies are built. The
    priority is based on the number of dependents --- we prefer to build
    packages with more dependents first.

    To build packages in parallel, we use a thread pool of n_jobs many
    threads listening to build_queue. When the thread is free, it takes an
    item off build_queue and builds it. Once the package is built, it sends the
    package to the built_queue. The main thread listens to the built_queue and
    checks if any of the dependents are ready to be built. If so, it adds the
    package to the build queue.
    """

    # Insert packages into build_queue. We *must* do this after counting
    # dependents, because the ordering ought not to change after insertion.

    if force_rebuild:
        # If "force_rebuild" is set, just rebuild everything
        needs_build = set(pkg_map.keys())
    else:
        needs_build = generate_needs_build_set(pkg_map, build_dir)

    # We won't rebuild the complement of the packages that we will build.
    already_built = set(pkg_map.keys()).difference(needs_build)

    # Remove the packages we've already built from the dependency sets of
    # the remaining ones
    for pkg_name in needs_build:
        pkg_map[pkg_name].unbuilt_host_dependencies.difference_update(already_built)

    needs_rust = any(
        pkg_map[pkg_name].meta.is_rust_package() for pkg_name in needs_build
    )
    if needs_rust:
        _ensure_rust_toolchain()

    if already_built:
        logger.info(
            "The following packages are already built: [bold]%s[/bold]",
            format_name_list(sorted(already_built)),
        )
    if not needs_build:
        logger.success("All packages already built. Quitting.")
        return

    sorted_needs_build = sorted(needs_build)
    logger.info(
        "Building the following packages: [bold]%s[/bold]",
        format_name_list(sorted_needs_build),
    )
    build_state = _GraphBuilder(pkg_map, build_args, build_dir, set(needs_build))
    try:
        build_state.run(n_jobs, already_built)
    except BuildError as err:
        logger.error(err.msg)
        sys.exit(err.returncode)


def generate_packagedata(
    output_dir: Path, pkg_map: dict[str, BasePackage]
) -> dict[str, PackageLockSpec]:
    packages: dict[str, PackageLockSpec] = {}
    for name, pkg in pkg_map.items():
        normalized_name = canonicalize_name(name)

        if not pkg.file_name or pkg.package_type == "static_library":
            continue

        wheel_file = output_dir / pkg.file_name
        if not wheel_file.exists():
            continue
        pkg_entry = PackageLockSpec(
            name=name,
            version=pkg.version,
            file_name=pkg.file_name,
            install_dir=pkg.install_dir,
            package_type=pkg.package_type,
        )

        pkg_entry.depends = [x.lower() for x in pkg.run_dependencies]

        if pkg.package_type not in ("static_library", "shared_library"):
            pkg_entry.imports = (
                pkg.meta.package.top_level if pkg.meta.package.top_level else [name]
            )

        if pkg.unvendor_tests:
            unvendored_test_file = unvendor.unvendor_tests_in_wheel(
                wheel_file, pkg.meta.build.retain_test_patterns
            )
            if unvendored_test_file:
                pkg_entry.unvendored_tests = True
                test_file_entry = PackageLockSpec(
                    name=f"{name}-tests",
                    version=pkg.version,
                    depends=[name],
                    file_name=unvendored_test_file.name,
                    install_dir=pkg.install_dir,
                )
                update_package_sha256(test_file_entry, unvendored_test_file)
                packages[normalized_name + "-tests"] = test_file_entry

        update_package_sha256(pkg_entry, wheel_file)
        packages[normalized_name] = pkg_entry

    # sort packages by name
    packages = dict(sorted(packages.items()))
    return packages


def generate_lockfile(
    output_dir: Path, pkg_map: dict[str, BasePackage]
) -> PyodideLockSpec:
    """Generate the package.json file"""

    # Build package.json data.
    [platform, _, arch] = build_env.platform().rpartition("_")
    info = {
        "arch": arch,
        "platform": platform,
        "version": build_env.get_build_flag("PYODIDE_VERSION"),
        "python": sys.version.partition(" ")[0],
        "abi_version": build_env.get_build_flag("PYODIDE_ABI_VERSION"),
    }
    packages = generate_packagedata(output_dir, pkg_map)
    lock_spec = PyodideLockSpec(info=info, packages=packages)
    lock_spec.check_wheel_filenames()
    return lock_spec


def copy_packages_to_dist_dir(
    packages: Iterable[BasePackage],
    output_dir: Path,
    compression_level: int = 6,
    metadata_files: bool = False,
) -> None:
    for pkg in packages:
        dist_artifact_path = pkg.dist_artifact_path()
        if dist_artifact_path:
            shutil.copy(dist_artifact_path, output_dir)
            repack_zip_archive(
                output_dir / dist_artifact_path.name,
                compression_level=compression_level,
            )

            if metadata_files and dist_artifact_path.suffix == ".whl":
                extract_wheel_metadata_file(
                    dist_artifact_path,
                    output_dir / f"{dist_artifact_path.name}.metadata",
                )


def build_packages(
    packages_dir: Path,
    targets: str,
    build_args: BuildArgs,
    build_dir: Path,
    n_jobs: int = 1,
    force_rebuild: bool = False,
) -> dict[str, BasePackage]:
    requested, disabled = _parse_package_query(targets)
    requested_packages = loader.load_recipes(packages_dir, requested)
    pkg_map = generate_dependency_graph(
        packages_dir, set(requested_packages.keys()), disabled
    )

    build_from_graph(pkg_map, build_args, build_dir, n_jobs, force_rebuild)
    for pkg in pkg_map.values():
        dist_path = pkg.dist_artifact_path()

        if dist_path:
            pkg.file_name = dist_path.name

    return pkg_map


def copy_logs(pkg_map: dict[str, BasePackage], log_dir: Path) -> None:
    """
    Copy build logs of packages to the log directory.
    Parameters
    ----------
    pkg_map
        A dictionary mapping package names to package objects.
    log_dir
        The directory to copy the logs to.
    """

    log_dir.mkdir(exist_ok=True, parents=True)
    logger.info("Copying build logs to %s", log_dir)

    for pkg in pkg_map.values():
        log_file = pkg.pkgdir / "build.log"
        if log_file.exists():
            shutil.copy(log_file, log_dir / f"{pkg.name}.log")
        else:
            logger.warning("Warning: %s has no build log", pkg.name)


def install_packages(
    pkg_map: dict[str, BasePackage],
    output_dir: Path,
    compression_level: int = 6,
    metadata_files: bool = False,
) -> None:
    """
    Install packages into the output directory.
    - copies build artifacts (wheel, zip, ...) to the output directory
    - create pyodide-lock.json


    pkg_map
        package map created from build_packages

    output_dir
        output directory to install packages into
    """

    output_dir.mkdir(exist_ok=True, parents=True)

    logger.info("Copying built packages to %s", output_dir)
    copy_packages_to_dist_dir(
        pkg_map.values(),
        output_dir,
        compression_level=compression_level,
        metadata_files=metadata_files,
    )

    lockfile_path = output_dir / "pyodide-lock.json"
    logger.info("Writing pyodide-lock.json to %s", lockfile_path)

    package_data = generate_lockfile(output_dir, pkg_map)
    package_data.to_json(lockfile_path)


def set_default_build_args(build_args: BuildArgs) -> BuildArgs:
    args = dataclasses.replace(build_args)

    if not args.cflags:
        args.cflags = build_env.get_build_flag("SIDE_MODULE_CFLAGS")
    if not args.cxxflags:
        args.cxxflags = build_env.get_build_flag("SIDE_MODULE_CXXFLAGS")
    if not args.ldflags:
        args.ldflags = build_env.get_build_flag("SIDE_MODULE_LDFLAGS")
    if not args.target_install_dir:
        args.target_install_dir = build_env.get_build_flag("TARGETINSTALLDIR")
    if not args.host_install_dir:
        args.host_install_dir = build_env.get_build_flag("HOSTINSTALLDIR")

    return args



from pathlib import Path

import typer

from pyodide_build.build_env import init_environment
from pyodide_build.out_of_tree import venv


# TODO: disabled options that can be later supported have been commented out, fix them
# --copies/--always-copy and symlink_app_data
def main(
    dest: Path = typer.Argument(
        ...,
        help="directory to create virtualenv at",
    ),
    clear: bool = typer.Option(
        False,
        "--clear/--no-clear",
        help="Remove the destination directory if it exists",
    ),
    no_vcs_ignore: bool = typer.Option(
        False,
        "--no-vcs-ignore",
        help="Don't create VCS ignore directive in the destination directory",
    ),
    # copies: bool = typer.Option(
    #     False, "--copies", "--always-copy", help="Use copies rather than symlinks, even when symlinks are the default"
    # ),
    no_download: bool = typer.Option(
        False,
        "--no-download",
        "--never-download",
        help="Disable download of the latest pip/setuptools from PyPI",
    ),
    download: bool = typer.Option(
        False,
        "--download/--no-download",
        help="Enable download of the latest pip/setuptools from PyPI",
    ),
    extra_search_dir: list[str] = typer.Option(
        None,
        "--extra-search-dir",
        help="A path containing wheels to extend the internal wheel list",
    ),
    pip: str = typer.Option(
        "bundle",
        "--pip",
        help="Version of pip to install as seed: embed, bundle, or exact version.",
    ),
    setuptools: str | None = typer.Option(
        None,
        "--setuptools",
        help="Version of setuptools to install as seed: embed, bundle, none or exact version",
    ),
    no_setuptools: bool = typer.Option(
        False, "--no-setuptools", help="Do not install setuptools"
    ),
    no_periodic_update: bool = typer.Option(
        False,
        "--no-periodic-update",
        help="Disable the periodic update of the embedded wheels",
    ),
    # symlink_app_data: bool = typer.Option(
    #     False, "--symlink-app-data/--no-symlink-app-data", help="Symlink the python packages from the app-data folder"
    # ),
) -> None:
    """
    Create a Pyodide virtual environment.
    Additionally, this interface supports a subset of the arguments that `virtualenv` supports, with some minor differences for Pyodide compatibility.
    Please note that passing extra options is experimental and may be subject to change.
    """
    init_environment()

    venv_args = []

    # if copies:
    #     venv_args.append("--copies")
    # if symlink_app_data:
    #     venv_args.append("--symlink-app-data")

    if clear:
        venv_args.append("--clear")
    if no_vcs_ignore:
        venv_args.append("--no-vcs-ignore")
    if no_download:
        venv_args.append("--no-download")
    if download:
        venv_args.append("--download")
    if extra_search_dir:
        for search_dir in extra_search_dir:
            venv_args.extend(["--extra-search-dir", search_dir])
    if pip:
        venv_args.extend(["--pip", pip])
    if setuptools is not None:
        venv_args.extend(["--setuptools", setuptools])
    if no_setuptools:
        venv_args.append("--no-setuptools")

    if no_periodic_update:
        venv_args.append("--no-periodic-update")

    venv.create_pyodide_venv(dest, venv_args)



import re
from pathlib import Path

import typer

from pyodide_build.pyzip import create_zipfile


def main(
    libdir: list[Path] = typer.Argument(
        ...,
        help="List of paths to the directory containing the Python standard library or extra packages.",
    ),
    exclude: str = typer.Option(
        "",
        help="List of files to exclude from the zip file. Defaults to no files.",
    ),
    stub: str = typer.Option(
        "",
        help="List of files that are replaced by JS implementations. Defaults to no files.",
    ),
    pycompile: bool = typer.Option(
        False, help="Whether to compile the .py files into .pyc."
    ),
    output: Path = typer.Option(
        "python.zip", help="Path to the output zip file. Defaults to python.zip."
    ),
    compression_level: int = typer.Option(
        6, help="Compression level to use for the created zip file"
    ),
) -> None:
    """
    Bundle Python standard libraries into a zip file.
    """

    # Convert the comma / space separated strings to lists
    excludes = [
        item.strip() for item in re.split(r",|\s", exclude) if item.strip() != ""
    ]
    stubs = [item.strip() for item in re.split(r",|\s", stub) if item.strip() != ""]

    create_zipfile(
        libdir,
        excludes,
        stubs,
        output,
        pycompile=pycompile,
        filterfunc=None,
        compression_level=compression_level,
    )
    typer.echo(f"Zip file created at {output.resolve()}")


main.typer_kwargs = {"hidden": True}  # type: ignore[attr-defined]



import re
import shutil
import sys
import tempfile
from pathlib import Path
from typing import cast, get_args
from urllib.parse import urlparse

import requests
import typer
from build import ConfigSettingsType

from pyodide_build.build_env import (
    check_emscripten_version,
    get_pyodide_root,
    init_environment,
)
from pyodide_build.common import default_xbuildenv_path
from pyodide_build.logger import logger
from pyodide_build.out_of_tree import build
from pyodide_build.out_of_tree.pypi import (
    build_dependencies_for_wheel,
    build_wheels_from_pypi_requirements,
    fetch_pypi_package,
)
from pyodide_build.pypabuild import parse_backend_flags
from pyodide_build.spec import _BuildSpecExports, _ExportTypes


def convert_exports(exports: str) -> _BuildSpecExports:
    if "," in exports:
        return [x.strip() for x in exports.split(",") if x.strip()]
    possible_exports = get_args(_ExportTypes)
    if exports in possible_exports:
        return cast(_ExportTypes, exports)
    logger.stderr(
        f"Expected exports to be one of "
        '"pyinit", "requested", "whole_archive", '
        "or a comma separated list of symbols to export. "
        f'Got "{exports}".'
    )
    sys.exit(1)


def pypi(
    package: str,
    output_directory: Path,
    exports: str,
    config_settings: ConfigSettingsType,
    isolation: bool = True,
    skip_dependency_check: bool = False,
) -> Path:
    """Fetch a wheel from pypi, or build from source if none available."""
    with tempfile.TemporaryDirectory() as tmpdir:
        srcdir = Path(tmpdir)

        # get package from pypi
        package_path = fetch_pypi_package(package, srcdir)
        if not package_path.is_dir():
            # a pure-python wheel has been downloaded - just copy to dist folder
            dest_file = output_directory / package_path.name
            shutil.copyfile(str(package_path), output_directory / package_path.name)
            print(f"Successfully fetched: {package_path.name}")
            return dest_file

        built_wheel = build.run(
            srcdir,
            output_directory,
            convert_exports(exports),
            config_settings,
            isolation=isolation,
            skip_dependency_check=skip_dependency_check,
        )
        return built_wheel


def download_url(url: str, output_directory: Path) -> str:
    with requests.get(url, stream=True) as response:
        urlpath = Path(urlparse(response.url).path)
        if urlpath.suffix == ".gz":
            urlpath = urlpath.with_suffix("")
        file_name = urlpath.name
        with open(output_directory / file_name, "wb") as f:
            for chunk in response.iter_content(chunk_size=1 << 20):
                f.write(chunk)
        return file_name


def url(
    package_url: str,
    output_directory: Path,
    exports: str,
    config_settings: ConfigSettingsType,
    isolation: bool = True,
    skip_dependency_check: bool = False,
) -> Path:
    """Fetch a wheel or build sdist from url."""
    with tempfile.TemporaryDirectory() as tmpdir:
        tmppath = Path(tmpdir)
        filename = download_url(package_url, tmppath)
        if Path(filename).suffix == ".whl":
            shutil.move(tmppath / filename, output_directory / filename)
            return output_directory / filename

        builddir = tmppath / "build"
        shutil.unpack_archive(tmppath / filename, builddir)
        files = list(builddir.iterdir())
        if len(files) == 1 and files[0].is_dir():
            # unzipped into subfolder
            builddir = files[0]
        wheel_path = build.run(
            builddir,
            output_directory,
            convert_exports(exports),
            config_settings,
            isolation=isolation,
            skip_dependency_check=skip_dependency_check,
        )
        return wheel_path


def source(
    source_location: Path,
    output_directory: Path,
    exports: str,
    config_settings: ConfigSettingsType,
    isolation: bool = True,
    skip_dependency_check: bool = False,
) -> Path:
    """Use pypa/build to build a Python package from source"""
    built_wheel = build.run(
        source_location,
        output_directory,
        convert_exports(exports),
        config_settings,
        isolation=isolation,
        skip_dependency_check=skip_dependency_check,
    )
    return built_wheel


DEFAULT_PATH = default_xbuildenv_path()


# simple 'pyodide build' command
def main(
    source_location: str | None = typer.Argument(
        "",
        help="Build source, can be source folder, pypi version specification, "
        "or url to a source dist archive or wheel file. If this is blank, it "
        "will build the current directory.",
    ),
    output_directory: str = typer.Option(
        "",
        "--outdir",
        "-o",
        help="which directory should the output be placed into?",
    ),
    requirements_txt: str = typer.Option(
        "",
        "--requirements",
        "-r",
        help="Build a list of package requirements from a requirements.txt file",
    ),
    exports: str = typer.Option(
        "requested",
        envvar="PYODIDE_BUILD_EXPORTS",
        help="Which symbols should be exported when linking .so files?",
    ),
    build_dependencies: bool = typer.Option(
        False, help="Fetch dependencies from pypi and build them too."
    ),
    output_lockfile: str = typer.Option(
        "",
        help="Output list of resolved dependencies to a file in requirements.txt format",
    ),
    skip_dependency: list[str] = typer.Option(
        [],
        help="Skip building or resolving a single dependency, or a pyodide-lock.json file. "
        "Use multiple times or provide a comma separated list to skip multiple dependencies.",
    ),
    skip_built_in_packages: bool = typer.Option(
        True,
        help="Don't build dependencies that are built into the pyodide distribution.",
    ),
    compression_level: int = typer.Option(
        6, help="Compression level to use for the created zip file"
    ),
    no_isolation: bool = typer.Option(
        False,
        "--no-isolation",
        "-n",
        help="Disable building the project in an isolated virtual environment. "
        "Build dependencies must be installed separately when this option is used",
    ),
    skip_dependency_check: bool = typer.Option(
        False,
        "--skip-dependency-check",
        "-x",
        help="Do not check that the build dependencies are installed. This option "
        "is only useful when used with --no-isolation.",
    ),
    config_setting: list[str] | None = typer.Option(
        None,
        "--config-setting",
        "-C",
        help=(
            "Settings to pass to the backend. "
            "Works same as the --config-setting option of pypa/build."
        ),
        metavar="KEY[=VALUE]",
    ),
    xbuildenv_path: Path = typer.Option(
        DEFAULT_PATH,
        "--xbuildenv-path",
        envvar="PYODIDE_XBUILDENV_PATH",
        help="Path to the cross-build environment directory.",
    ),
    ctx: typer.Context = typer.Context,  # type: ignore[assignment]
) -> None:
    """Use pypa/build to build a Python package from source, pypi or url."""
    init_environment(xbuildenv_path=xbuildenv_path)
    try:
        check_emscripten_version()
    except RuntimeError as e:
        print(e.args[0], file=sys.stderr)
        sys.exit(1)

    output_directory = output_directory or "./dist"

    outpath = Path(output_directory).resolve()
    outpath.mkdir(exist_ok=True)
    extras: list[str] = []

    # For backward compatibility, in addition to the `--config-setting` arguments, we also support
    # passing config settings as positional arguments.
    config_settings = parse_backend_flags((config_setting or []) + ctx.args)

    if skip_built_in_packages:
        package_lock_json = get_pyodide_root() / "dist" / "pyodide-lock.json"
        skip_dependency.append(str(package_lock_json.absolute()))

    if len(requirements_txt) > 0:
        # a requirements.txt - build it (and optionally deps)
        if not Path(requirements_txt).exists():
            raise RuntimeError(
                f"Couldn't find requirements text file {requirements_txt}"
            )
        reqs = []
        with open(requirements_txt) as f:
            raw_reqs = [x.strip() for x in f.readlines()]
        for x in raw_reqs:
            # remove comments
            comment_pos = x.find("#")
            if comment_pos != -1:
                x = x[:comment_pos].strip()
            if len(x) > 0:
                if x[0] == "-":
                    raise RuntimeError(
                        f"pyodide build only supports name-based PEP508 requirements. [{x}] will not work."
                    )
                if x.find("@") != -1:
                    raise RuntimeError(
                        f"pyodide build does not support URL based requirements. [{x}] will not work"
                    )
                reqs.append(x)
        try:
            build_wheels_from_pypi_requirements(
                reqs,
                outpath,
                build_dependencies,
                skip_dependency,
                # TODO: should we really use same "exports" value for all of our
                # dependencies? Not sure this makes sense...
                convert_exports(exports),
                config_settings,
                isolation=not no_isolation,
                skip_dependency_check=skip_dependency_check,
                output_lockfile=output_lockfile,
            )
        except BaseException as e:
            import traceback

            print("Failed building multiple wheels:", traceback.format_exc())
            raise e
        return

    if source_location is not None:
        extras = re.findall(r"\[(\w+)\]", source_location)
        if len(extras) != 0:
            source_location = source_location[0 : source_location.find("[")]
    if not source_location:
        # build the current folder
        wheel = source(
            Path.cwd(),
            outpath,
            exports,
            config_settings,
            isolation=not no_isolation,
            skip_dependency_check=skip_dependency_check,
        )
    elif source_location.find("://") != -1:
        wheel = url(
            source_location,
            outpath,
            exports,
            config_settings,
            isolation=not no_isolation,
            skip_dependency_check=skip_dependency_check,
        )
    elif Path(source_location).is_dir():
        # a folder, build it
        wheel = source(
            Path(source_location).resolve(),
            outpath,
            exports,
            config_settings,
            isolation=not no_isolation,
            skip_dependency_check=skip_dependency_check,
        )
    elif source_location.find("/") == -1:
        # try fetch or build from pypi
        wheel = pypi(
            source_location,
            outpath,
            exports,
            config_settings,
            isolation=not no_isolation,
            skip_dependency_check=skip_dependency_check,
        )
    else:
        raise RuntimeError(f"Couldn't determine source type for {source_location}")
    # now build deps for wheel
    if build_dependencies:
        try:
            build_dependencies_for_wheel(
                wheel,
                extras,
                skip_dependency,
                # TODO: should we really use same "exports" value for all of our
                # dependencies? Not sure this makes sense...
                convert_exports(exports),
                config_settings,
                isolation=not no_isolation,
                skip_dependency_check=skip_dependency_check,
                output_lockfile=output_lockfile,
                compression_level=compression_level,
            )
        except BaseException as e:
            import traceback

            print("Failed building dependencies for wheel:", traceback.format_exc())
            wheel.unlink()
            raise e


main.typer_kwargs = {  # type: ignore[attr-defined]
    "context_settings": {
        "ignore_unknown_options": True,
        "allow_extra_args": True,
        "help_option_names": ["-h", "--help"],
    },
}



import logging
import os
from collections.abc import Generator
from contextlib import contextmanager

from rich.console import Console
from rich.highlighter import NullHighlighter
from rich.logging import RichHandler
from rich.theme import Theme

IN_CI = "CI" in os.environ
IN_PYTEST = "IN_PYTEST" in os.environ

COLOR_THEME = Theme(
    {
        "debug": "",
        "info": "",
        "warning": "bold yellow",
        "error": "bold red",
        "critical": "bold red",
    }
)

LEVEL_STDOUT = logging.INFO - 5
LEVEL_STDERR = logging.INFO + 5


# Note: it seems like rich's auto terminal detection mechanism doesn't work well for CircleCI.
#       so we manually disable it when running in CI.
class CIAwareConsole(Console):
    @property
    def is_terminal(self) -> bool:
        """Check if the console is writing to a terminal."""
        return not IN_CI and not IN_PYTEST and super().is_terminal


class StdoutFilter(logging.Filter):
    def filter(self, record):
        return record.levelno in (LEVEL_STDOUT, logging.DEBUG, logging.INFO)


class StderrFilter(logging.Filter):
    def filter(self, record):
        return record.levelno in (
            LEVEL_STDERR,
            logging.WARNING,
            logging.ERROR,
            logging.CRITICAL,
        )


class _Logger(logging.Logger):
    def success(self, msg, *args, **kwargs):
        self.info("[bold green]%s[/bold green]", msg, *args, **kwargs)

    def stdout(self, msg, *args, **kwargs):
        self.log(LEVEL_STDOUT, msg, *args, **kwargs)

    def stderr(self, msg, *args, **kwargs):
        self.log(LEVEL_STDERR, msg, *args, **kwargs)


class RichFormatter(logging.Formatter):
    """Colorize log messages based on log level."""

    def format(self, records: logging.LogRecord) -> str:
        levelname = records.levelname.lower()
        records.msg = f"[{levelname}]{records.msg}[/{levelname}]"

        formatted = super().format(records)
        return formatted


console_stdout = CIAwareConsole(theme=COLOR_THEME)
console_stderr = CIAwareConsole(stderr=True, theme=COLOR_THEME)


def _get_logger(log_level: int) -> _Logger:
    logger = _Logger(__name__)
    logger.setLevel(log_level)

    rich_handler_stdout = RichHandler(
        console=console_stdout,
        highlighter=NullHighlighter(),
        show_time=False,
        show_level=False,
        show_path=False,
        markup=True,
    )
    rich_handler_stdout.setFormatter(RichFormatter("%(message)s"))
    rich_handler_stderr = RichHandler(
        console=console_stderr,
        highlighter=NullHighlighter(),
        show_time=False,
        show_level=False,
        show_path=False,
        markup=True,
    )
    rich_handler_stderr.setFormatter(RichFormatter("%(message)s"))

    rich_handler_stdout.addFilter(StdoutFilter())
    rich_handler_stderr.addFilter(StderrFilter())

    logger.addHandler(rich_handler_stdout)
    logger.addHandler(rich_handler_stderr)

    return logger


logger = _get_logger(logging.INFO)


@contextmanager
def set_log_level(
    _logger: logging.Logger, log_level: int | bool
) -> Generator[None, None, None]:
    """Set the log level.

    Parameters
    ----------
    logger
        The logger to set the log level for.

    log_level
        If True, set the log level to DEBUG (verbose mode).
        If given an integer, set the log level to that value.
    """

    original_log_level = logger.level

    if isinstance(log_level, bool):
        log_level = logging.DEBUG if log_level else original_log_level

    _logger.setLevel(log_level)

    yield

    _logger.setLevel(original_log_level)


if __name__ == "__main__":
    # For testing the colors
    logger.debug("debug")
    logger.info("info")
    logger.stdout("stdout")
    logger.warning("warning")
    logger.error("error")
    logger.success("success")
    logger.stderr("stderr")



import dataclasses
import sys
from pathlib import Path

import typer

from pyodide_build import build_env
from pyodide_build.build_env import BuildArgs, init_environment
from pyodide_build.common import get_num_cores
from pyodide_build.logger import logger
from pyodide_build.recipe import graph_builder, loader
from pyodide_build.recipe.builder import RecipeBuilder


@dataclasses.dataclass(eq=False, order=False, kw_only=True)
class Args:
    recipe_dir: Path
    build_dir: Path
    install_dir: Path
    build_args: BuildArgs
    skip_rust_setup: bool
    force_rebuild: bool
    n_jobs: int

    def __init__(
        self,
        *,
        recipe_dir: Path | str | None,
        build_dir: Path | str | None,
        install_dir: Path | str | None = None,
        build_args: BuildArgs,
        force_rebuild: bool,
        skip_rust_setup: bool = False,
        n_jobs: int | None = None,
    ):
        cwd = Path.cwd()
        root = build_env.search_pyodide_root(cwd) or cwd
        self.recipe_dir = (
            root / "packages" if not recipe_dir else Path(recipe_dir).resolve()
        )
        self.build_dir = self.recipe_dir if not build_dir else Path(build_dir).resolve()
        self.install_dir = (
            root / "dist" if not install_dir else Path(install_dir).resolve()
        )
        self.build_args = build_args
        self.force_rebuild = force_rebuild
        self.skip_rust_setup = skip_rust_setup
        self.n_jobs = n_jobs or get_num_cores()
        if not self.recipe_dir.is_dir():
            raise FileNotFoundError(f"Recipe directory {self.recipe_dir} not found")


@dataclasses.dataclass(eq=False, order=False, kw_only=True)
class InstallOptions:
    compression_level: int
    metadata_files: bool


def build_recipes_no_deps(
    packages: list[str] = typer.Argument(
        ..., help="Packages to build, or ``*`` for all packages in recipe directory"
    ),
    recipe_dir: str = typer.Option(
        None,
        help="The directory containing the recipe of packages. "
        "If not specified, the default is ``./packages``",
    ),
    build_dir: str = typer.Option(
        None,
        envvar="PYODIDE_RECIPE_BUILD_DIR",
        help="The directory where build directories for packages are created. "
        "Default: recipe_dir.",
    ),
    cflags: str = typer.Option(
        "", help="Extra compiling flags. Default: SIDE_MODULE_CFLAGS"
    ),
    cxxflags: str = typer.Option(
        "", help="Extra compiling flags. Default: SIDE_MODULE_CXXFLAGS"
    ),
    ldflags: str = typer.Option(
        "", help="Extra linking flags. Default: SIDE_MODULE_LDFLAGS"
    ),
    target_install_dir: str = typer.Option(
        "",
        help="The path to the target Python installation. Default: TARGETINSTALLDIR",
    ),
    host_install_dir: str = typer.Option(
        "",
        help="Directory for installing built host packages. Default: HOSTINSTALLDIR",
    ),
    force_rebuild: bool = typer.Option(
        False,
        help="Force rebuild of all packages regardless of whether they appear to have been updated",
    ),
    continue_: bool = typer.Option(
        False,
        "--continue",
        help="Continue a build from the middle. For debugging. Implies '--force-rebuild'",
    ),
    skip_rust_setup: bool = typer.Option(
        False,
        "--skip-rust-setup",
        help="Don't setup rust environment when building a rust package",
    ),
) -> None:
    """Build packages using yaml recipes but don't try to resolve dependencies"""
    init_environment()

    if build_env.in_xbuildenv():
        build_env.check_emscripten_version()

    build_args = BuildArgs(
        cflags=cflags,
        cxxflags=cxxflags,
        ldflags=ldflags,
        target_install_dir=target_install_dir,
        host_install_dir=host_install_dir,
    )
    build_args = graph_builder.set_default_build_args(build_args)
    args = Args(
        build_args=build_args,
        build_dir=build_dir,
        recipe_dir=recipe_dir,
        force_rebuild=force_rebuild,
        skip_rust_setup=skip_rust_setup,
    )

    return build_recipes_no_deps_impl(packages, args, continue_)


def _rust_setup(recipe_dir: Path, packages: list[str]) -> None:
    recipes = loader.load_recipes(recipe_dir, packages, False)
    if any(recipe.is_rust_package() for recipe in recipes.values()):
        graph_builder._ensure_rust_toolchain()


def build_recipes_no_deps_impl(
    packages: list[str], args: Args, continue_: bool
) -> None:
    # TODO: use multiprocessing?
    if not args.skip_rust_setup:
        _rust_setup(args.recipe_dir, packages)

    for package in packages:
        package_path = args.recipe_dir / package
        package_build_dir = args.build_dir / package / "build"
        builder = RecipeBuilder.get_builder(
            package_path,
            args.build_args,
            package_build_dir,
            args.force_rebuild,
            continue_,
        )
        builder.build()


def build_recipes(
    packages: list[str] = typer.Argument(
        ..., help="Packages to build, or ``*`` for all packages in recipe directory"
    ),
    recipe_dir: str = typer.Option(
        None,
        help="The directory containing the recipe of packages. "
        "If not specified, the default is ``./packages``",
    ),
    build_dir: str = typer.Option(
        None,
        envvar="PYODIDE_RECIPE_BUILD_DIR",
        help="The directory where build directories for packages are created. "
        "Default: recipe_dir.",
    ),
    install: bool = typer.Option(
        False,
        help="If true, install the built packages into the install_dir. "
        "If false, build packages without installing.",
    ),
    install_dir: str = typer.Option(
        None,
        help="Path to install built packages and pyodide-lock.json. "
        "If not specified, the default is ``./dist``.",
    ),
    metadata_files: bool = typer.Option(
        False,
        help="If true, extract the METADATA file from the built wheels "
        "to a matching ``*.whl.metadata`` file. "
        "If false, no ``*.whl.metadata`` file is produced.",
    ),
    no_deps: bool = typer.Option(
        False, help="Removed, use `pyodide build-recipes-no-deps` instead."
    ),
    cflags: str = typer.Option(
        None, help="Extra compiling flags. Default: SIDE_MODULE_CFLAGS"
    ),
    cxxflags: str = typer.Option(
        None, help="Extra compiling flags. Default: SIDE_MODULE_CXXFLAGS"
    ),
    ldflags: str = typer.Option(
        None, help="Extra linking flags. Default: SIDE_MODULE_LDFLAGS"
    ),
    target_install_dir: str = typer.Option(
        "",
        help="The path to the target Python installation. Default: TARGETINSTALLDIR",
    ),
    host_install_dir: str = typer.Option(
        "",
        help="Directory for installing built host packages. Default: HOSTINSTALLDIR",
    ),
    log_dir: str = typer.Option(None, help="Directory to place log files"),
    force_rebuild: bool = typer.Option(
        False,
        help="Force rebuild of all packages regardless of whether they appear to have been updated",
    ),
    n_jobs: int = typer.Option(
        None,
        help="Number of packages to build in parallel  (default: # of cores in the system)",
    ),
    compression_level: int = typer.Option(
        6,
        envvar="PYODIDE_ZIP_COMPRESSION_LEVEL",
        help="Level of zip compression to apply when installing. 0 means no compression.",
    ),
) -> None:
    if no_deps:
        logger.error(
            "--no-deps has been removed, use pyodide build-package-no-deps instead",
        )
        sys.exit(1)
    if metadata_files and not install:
        logger.warning(
            "WARNING: when --install is not set, the --metadata-files parameter is ignored",
        )

    install_options: InstallOptions | None = None
    if install:
        install_options = InstallOptions(
            metadata_files=metadata_files, compression_level=compression_level
        )

    init_environment()

    if build_env.in_xbuildenv():
        build_env.check_emscripten_version()

    build_args = BuildArgs(
        cflags=cflags,
        cxxflags=cxxflags,
        ldflags=ldflags,
        target_install_dir=target_install_dir,
        host_install_dir=host_install_dir,
    )
    build_args = graph_builder.set_default_build_args(build_args)
    args = Args(
        build_args=build_args,
        build_dir=build_dir,
        install_dir=install_dir,
        recipe_dir=recipe_dir,
        force_rebuild=force_rebuild,
        n_jobs=n_jobs,
    )
    log_dir_ = Path(log_dir).resolve() if log_dir else None
    build_recipes_impl(packages, args, log_dir_, install_options)


def build_recipes_impl(
    packages: list[str],
    args: Args,
    log_dir: Path | None,
    install_options: InstallOptions | None,
) -> None:
    if len(packages) == 1 and "," in packages[0]:
        # Handle packages passed with old comma separated syntax.
        # This is to support `PYODIDE_PACKAGES="pkg1,pkg2,..." make` syntax.
        targets = packages[0].replace(" ", "")
    else:
        targets = ",".join(packages)

    pkg_map = graph_builder.build_packages(
        args.recipe_dir,
        targets=targets,
        build_args=args.build_args,
        build_dir=args.build_dir,
        n_jobs=args.n_jobs,
        force_rebuild=args.force_rebuild,
    )

    if log_dir:
        graph_builder.copy_logs(pkg_map, log_dir)

    if install_options:
        graph_builder.install_packages(
            pkg_map,
            args.install_dir,
            compression_level=install_options.compression_level,
            metadata_files=install_options.metadata_files,
        )



from pathlib import Path
from typing import Literal, Self

import pydantic
from pydantic import BaseModel, ConfigDict, Field

from pyodide_build.spec import _BuildSpecExports


class _PackageSpec(BaseModel):
    name: str
    version: str
    top_level: list[str] = Field([], alias="top-level")
    tag: list[str] = Field([])
    disabled: bool = Field(False, alias="_disabled")
    pinned: bool = Field(False, alias="pinned")
    model_config = ConfigDict(extra="forbid")


class _SourceSpec(BaseModel):
    url: str | None = None
    extract_dir: str | None = None
    path: Path | None = None
    sha256: str | None = None
    patches: list[str] = []
    extras: list[tuple[str, str]] = []
    model_config = ConfigDict(extra="forbid")

    @pydantic.model_validator(mode="after")
    def _check_url_has_hash(self) -> Self:
        if self.url is not None and self.sha256 is None:
            raise ValueError(
                "If source is downloaded from url, it must have a 'source/sha256' hash."
            )

        return self

    @pydantic.model_validator(mode="after")
    def _check_in_tree_url(self) -> Self:
        in_tree = self.path is not None
        from_url = self.url is not None

        # cpython_modules is a special case, it is not in the tree
        # TODO: just copy the file into the tree?
        # if not (in_tree or from_url):
        #     raise ValueError("Source section should have a 'url' or 'path' key")

        if in_tree and from_url:
            raise ValueError(
                "Source section should not have both a 'url' and a 'path' key"
            )

        return self

    @pydantic.model_validator(mode="after")
    def _check_patches_extra(self) -> Self:
        in_tree = self.path is not None
        url_is_wheel = self.url and self.url.endswith(".whl")

        if in_tree and (self.patches or self.extras):
            raise ValueError(
                "If source is in tree, 'source/patches' and 'source/extras' keys "
                "are not allowed"
            )

        if url_is_wheel and (self.patches or self.extras):
            raise ValueError(
                "If source is a wheel, 'source/patches' and 'source/extras' "
                "keys are not allowed"
            )

        return self


_BuildSpecTypes = Literal[
    "package", "static_library", "shared_library", "cpython_module"
]


class _BuildSpec(BaseModel):
    exports: _BuildSpecExports = "pyinit"
    backend_flags: str = Field("", alias="backend-flags")
    cflags: str = ""
    cxxflags: str = ""
    ldflags: str = ""
    package_type: _BuildSpecTypes = Field("package", alias="type")
    cross_script: str | None = Field(None, alias="cross-script")
    script: str | None = None
    post: str | None = None
    unvendor_tests: bool = Field(True, alias="unvendor-tests")
    retain_test_patterns: list[str] = Field([], alias="_retain-test-patterns")
    vendor_sharedlib: bool = Field(False, alias="vendor-sharedlib")
    cross_build_env: bool = Field(False, alias="cross-build-env")
    cross_build_files: list[str] = Field([], alias="cross-build-files")
    model_config = ConfigDict(extra="forbid")

    @pydantic.model_validator(mode="after")
    def _check_config(self) -> Self:
        static_library = self.package_type == "static_library"
        shared_library = self.package_type == "shared_library"
        cpython_module = self.package_type == "cpython_module"

        if not (static_library or shared_library or cpython_module):
            return self

        allowed_keys = {
            "package_type",
            "script",
            "exports",
            "unvendor_tests",
        }

        typ = self.package_type
        for key in self.model_fields_set:
            if key not in allowed_keys:
                raise ValueError(
                    f"If building a {typ}, 'build/{key}' key is not allowed."
                )

        return self


class _RequirementsSpec(BaseModel):
    run: list[str] = []
    host: list[str] = []
    executable: list[str] = []
    constraint: list[str] = []
    model_config = ConfigDict(extra="forbid")


class _TestSpec(BaseModel):
    imports: list[str] = []
    model_config = ConfigDict(extra="forbid")


class _AboutSpec(BaseModel):
    home: str | None = None
    PyPI: str | None = None
    summary: str | None = None
    license: str | None = None
    model_config = ConfigDict(extra="forbid")


class _ExtraSpec(BaseModel):
    recipe_maintainers: list[str] = Field([], alias="recipe-maintainers")


class MetaConfig(BaseModel):
    package: _PackageSpec
    source: _SourceSpec = _SourceSpec()
    build: _BuildSpec = _BuildSpec()
    requirements: _RequirementsSpec = _RequirementsSpec()
    test: _TestSpec = _TestSpec()
    about: _AboutSpec = _AboutSpec()
    extra: _ExtraSpec = _ExtraSpec()
    model_config = ConfigDict(extra="forbid")

    @classmethod
    def from_yaml(cls, path: Path) -> "MetaConfig":
        """Load the meta.yaml from a path

        Parameters
        ----------
        path
            path to the meta.yaml file
        """
        from ruamel.yaml import YAML

        yaml = YAML(typ="safe")

        config_raw = yaml.load(path)

        config = cls(**config_raw)
        if config.source.path:
            config.source.path = path.parent / config.source.path
        return config

    def to_yaml(self, path: Path) -> None:
        """Serialize the configuration to meta.yaml file

        Parameters
        ----------
        path
            path to the meta.yaml file
        """
        from ruamel.yaml import YAML

        yaml = YAML()
        yaml.representer.ignore_aliases = lambda *_: True

        with open(path, "w") as f:
            yaml.dump(self.model_dump(by_alias=True, exclude_unset=True), f)

    @pydantic.model_validator(mode="after")
    def _check_wheel_host_requirements(self) -> Self:
        """Check that if sources is a wheel it shouldn't have host dependencies."""
        if self.source.path is None and self.source.url is None:
            raise ValueError(
                'either "path" or "url" must be provided in the "source" section'
            )

        source_url = self.source.url
        requirements_host = self.requirements.host

        if source_url is not None and source_url.endswith(".whl"):
            if len(requirements_host):
                raise ValueError(
                    f"When source -> url is a wheel ({source_url}) the package cannot have host "
                    f"dependencies. Found {requirements_host}"
                )

            allowed_keys = {
                "post",
                "unvendor-tests",
                # Note here names are with "_", after alias conversion
                "cross_build_env",
                "cross_build_files",
                "exports",
                "unvendor_tests",
                "package_type",
            }
            for key in self.build.model_fields_set:
                if key not in allowed_keys:
                    raise ValueError(
                        f"If source is a wheel, 'build/{key}' key is not allowed"
                    )

        return self

    def is_rust_package(self) -> bool:
        """
        Check if a package requires rust toolchain to build.
        """
        return any(
            q in self.requirements.executable for q in ("rustc", "cargo", "rustup")
        )



# Common functions shared by other modules.
# Notes for contributors:
#   This module should not import any other modules from pyodide-build except logger to avoid circular imports.

import contextlib
import hashlib
import os
import shutil
import subprocess
import sys
import textwrap
import time
import tomllib
import warnings
import zipfile
from collections import deque
from collections.abc import Generator, Iterable, Iterator, Mapping, Sequence
from contextlib import contextmanager
from functools import cache
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, NoReturn
from urllib.request import urlopen
from zipfile import ZipFile

import platformdirs
from packaging.tags import Tag
from packaging.utils import canonicalize_name as canonicalize_package_name
from packaging.utils import parse_wheel_filename

from pyodide_build.logger import logger


def xbuildenv_dirname() -> str:
    try:
        from pyodide_build import __version__
    except ImportError:
        __version__ = "0.0.0"

    return f".pyodide-xbuildenv-{__version__}"


@cache
def default_xbuildenv_path() -> Path:
    """
    Return the default path to the cross-build environment directory. This directory
    is used when no path is provided to `pyodide xbuildenv`'s subcommands.

    The search order for determining the path is as follows:
    1. the PYODIDE_XBUILDENV_PATH environment variable if set
    2. the value provided by "pyodide config xbuildenv_path"
    3. the default location based on platformdirs

    Returns
    -------
    Path
        The path to the cross-build environment directory
    """

    # 1. check "pyodide config xbuildenv_path"
    try:
        from pyodide_build.build_env import get_host_build_flag

        config_path_str = get_host_build_flag("PYODIDE_XBUILDENV_PATH")

        # we can skip to fallback options if the config path is empty
        # as Path("") returns Path("."), which is not what we want.
        if config_path_str:
            config_path = Path(config_path_str).resolve()

            if _has_write_access(config_path):
                return config_path
            else:
                logger.error(
                    "Error: the directory specified in pyproject.toml or PYODIDE_XBUILDENV_PATH (%s) is not writable. ",
                    config_path,
                )

    except Exception as e:
        logger.error("Error reading xbuildenv_path from config system: %s", e)

    # 2. use default locations from platformdirs and elsewhere
    dirname = xbuildenv_dirname()
    candidates = []

    # 2.1. default cache directory
    candidates.append(Path(platformdirs.user_cache_dir()) / dirname)
    # 2.2. current working directory
    candidates.append(Path.cwd() / dirname)

    for candidate in candidates:
        if _has_write_access(candidate):
            return candidate

    raise RuntimeError(
        "Cannot find a writable directory for the cross-build environment. "
        "Please check if you have write access to the following directories:\n"
        f"  {', '.join(str(c) for c in candidates)}\n"
    )


def _has_write_access(folder: Path) -> bool:
    """
    Checks if the current user has write access to the given folder using pathlib.
    """
    try:
        if not folder.exists() and folder.parent != folder:
            return _has_write_access(folder.parent)

        return os.access(str(folder), os.W_OK)
    except OSError:
        return False


def _find_matching_wheels(
    wheel_paths: Iterable[Path],
    supported_tags: Sequence[Tag],
    version: str | None = None,
) -> Iterator[Path]:
    """
    Returns the sequence wheels whose tags match the Pyodide interpreter.

    We don't bother ordering them carefully because we are only hoping to find one.

    Parameters
    ----------
    wheel_paths
        A list of paths to wheels
    supported_tags
        A list of tags that the environment supports

    Returns
    -------
    The subset of wheel_paths that have tags that match the Pyodide interpreter.
    """
    for wheel_path in wheel_paths:
        _, wheel_version, _, wheel_tags = parse_wheel_filename(wheel_path.name)
        if version and version != str(wheel_version):
            continue
        for supported_tag in supported_tags:
            if supported_tag in wheel_tags:
                yield wheel_path
                continue


def find_matching_wheel(
    wheel_paths: Iterable[Path], supported_tags: Sequence[Tag], version: str = None
) -> Path | None:
    """
    Find a matching wheel or raise an error if none is present.

    Parameters
    ----------
    wheel_paths
        A list of paths to wheels
    supported_tags
        A list of tags that the environment supports

    Returns
    -------
    The subset of wheel_paths that have tags that match the Pyodide interpreter.
    """
    result = list(_find_matching_wheels(wheel_paths, supported_tags, version))
    if not result:
        return None
    if len(result) > 1:
        raise RuntimeError(
            "Found multiple matching wheels:\n" + "\n".join(w.name for w in result)
        )
    return result[0]


def parse_top_level_import_name(whlfile: Path) -> list[str] | None:
    """
    Parse the top-level import names from a wheel file.
    """

    if not whlfile.name.endswith(".whl"):
        raise RuntimeError(f"{whlfile} is not a wheel file.")

    whlzip = zipfile.Path(whlfile)

    def _valid_package_name(dirname: str) -> bool:
        return all(invalid_chr not in dirname for invalid_chr in ".- ")

    def _has_python_file(subdir: zipfile.Path) -> bool:
        queue = deque([subdir])
        while queue:
            nested_subdir = queue.pop()
            for subfile in nested_subdir.iterdir():
                if subfile.is_file() and subfile.name.endswith(".py"):
                    return True
                elif subfile.is_dir() and _valid_package_name(subfile.name):
                    queue.append(subfile)

        return False

    # If there is no top_level.txt file, we will find top level imports by
    # 1) a python file on a top-level directory
    # 2) a sub directory with __init__.py
    # following: https://github.com/pypa/setuptools/blob/d680efc8b4cd9aa388d07d3e298b870d26e9e04b/setuptools/discovery.py#L122
    top_level_imports = []
    for subdir in whlzip.iterdir():
        if subdir.is_file() and subdir.name.endswith(".py"):
            top_level_imports.append(subdir.name[:-3])
        elif subdir.is_dir() and _valid_package_name(subdir.name):
            if _has_python_file(subdir):
                top_level_imports.append(subdir.name)

    if not top_level_imports:
        logger.warning(
            "WARNING: failed to parse top level import name from %s.", whlfile
        )
        return None

    return top_level_imports


def _environment_substitute_str(
    string: str, env: Mapping[str, str] | None = None
) -> str:
    """
    Substitute $(VAR) in string with the value of the environment variable VAR.

    Parameters
    ----------
    string
        A string

    env
        A dictionary of environment variables. If None, use os.environ.

    Returns
    -------
    A string with the substitutions applied.
    """
    if env is None:
        env = dict(os.environ)

    for e_name, e_value in env.items():
        string = string.replace(f"$({e_name})", e_value)

    return string


def environment_substitute_args(
    args: dict[str, str], env: dict[str, str] | None = None
) -> dict[str, Any]:
    """
    Substitute $(VAR) in args with the value of the environment variable VAR.

    Parameters
    ----------
    args
        A dictionary of arguments

    env
        A dictionary of environment variables. If None, use os.environ.

    Returns
    -------
    A dictionary of arguments with the substitutions applied.
    """
    if env is None:
        env = dict(os.environ)
    subbed_args = {}
    for arg, value in args.items():
        if isinstance(value, str):
            value = _environment_substitute_str(value, env)
        subbed_args[arg] = value
    return subbed_args


@contextlib.contextmanager
def replace_env(build_env: Mapping[str, str]) -> Generator[None, None, None]:
    old_environ = dict(os.environ)
    os.environ.clear()
    os.environ.update(build_env)
    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(old_environ)


def exit_with_stdio(result: subprocess.CompletedProcess[str]) -> NoReturn:
    if result.stdout:
        logger.error("  stdout:")
        logger.error(textwrap.indent(result.stdout, "    "))
    if result.stderr:
        logger.error("  stderr:")
        logger.error(textwrap.indent(result.stderr, "    "))
    raise SystemExit(result.returncode)


def find_missing_executables(executables: list[str]) -> list[str]:
    return list(filter(lambda exe: shutil.which(exe) is None, executables))


@contextmanager
def chdir(new_dir: Path) -> Generator[None, None, None]:
    orig_dir = Path.cwd()
    try:
        os.chdir(new_dir)
        yield
    finally:
        os.chdir(orig_dir)


def get_num_cores() -> int:
    """
    Return the number of CPUs the current process can use.
    If the number of CPUs cannot be determined, return 1.
    """
    from pyodide_build.vendor.loky import cpu_count

    return cpu_count()


def make_zip_archive(
    archive_path: Path,
    input_dir: Path,
    compression_level: int = 6,
) -> None:
    """Create a zip archive out of a input folder

    Parameters
    ----------
    archive_path
       Path to the zip file that will be created
    input_dir
       input dir to compress
    compression_level
       compression level of the resulting zip file.
    """
    if compression_level > 0:
        compression = zipfile.ZIP_DEFLATED
    else:
        compression = zipfile.ZIP_STORED

    with zipfile.ZipFile(
        archive_path, "w", compression=compression, compresslevel=compression_level
    ) as zf:
        for file in input_dir.rglob("*"):
            zf.write(file, file.relative_to(input_dir))


def repack_zip_archive(archive_path: Path, compression_level: int = 6) -> None:
    """Repack zip archive with a different compression level"""
    if compression_level > 0:
        compression = zipfile.ZIP_DEFLATED
    else:
        compression = zipfile.ZIP_STORED

    with TemporaryDirectory() as temp_dir:
        input_path = Path(temp_dir) / archive_path.name
        shutil.move(archive_path, input_path)
        with (
            zipfile.ZipFile(input_path) as fh_zip_in,
            zipfile.ZipFile(
                archive_path,
                "w",
                compression=compression,
                compresslevel=compression_level,
            ) as fh_zip_out,
        ):
            for name in fh_zip_in.namelist():
                fh_zip_out.writestr(name, fh_zip_in.read(name))


def _get_sha256_checksum(archive: Path) -> str:
    """Compute the sha256 checksum of a file

    Parameters
    ----------
    archive
        the path to the archive we wish to checksum

    Returns
    -------
    checksum
         sha256 checksum of the archive
    """
    CHUNK_SIZE = 1 << 16
    h = hashlib.sha256()
    with open(archive, "rb") as fd:
        while True:
            chunk = fd.read(CHUNK_SIZE)
            h.update(chunk)
            if len(chunk) < CHUNK_SIZE:
                break
    return h.hexdigest()


def _format_dep_chain(dep_chain: Sequence[str]) -> str:
    return " -> ".join(dep.partition(";")[0].strip() for dep in dep_chain)


def _format_missing_dependencies(missing: set[tuple[str, ...]]) -> str:
    return "".join(
        "\n\t" + dep
        for deps in missing
        for dep in (deps[0], _format_dep_chain(deps[1:]))
        if dep
    )


def unpack_wheel(
    wheel_path: Path, target_dir: Path | None = None, verbose=True
) -> None:
    if target_dir is None:
        target_dir = wheel_path.parent
    result = subprocess.run(
        [sys.executable, "-m", "wheel", "unpack", wheel_path, "-d", target_dir],
        check=False,
        encoding="utf-8",
        capture_output=not verbose,
    )
    if result.returncode != 0:
        logger.error("ERROR: Unpacking wheel %s failed", wheel_path.name)
        exit_with_stdio(result)


def pack_wheel(wheel_dir: Path, target_dir: Path | None = None, verbose=True) -> None:
    if target_dir is None:
        target_dir = wheel_dir.parent
    result = subprocess.run(
        [sys.executable, "-m", "wheel", "pack", wheel_dir, "-d", target_dir],
        check=False,
        encoding="utf-8",
        capture_output=not verbose,
    )
    if result.returncode != 0:
        logger.error("ERROR: Packing wheel %s failed", wheel_dir)
        exit_with_stdio(result)


@contextmanager
def modify_wheel(wheel: Path, verbose=True) -> Iterator[Path]:
    """Unpacks the wheel into a temp directory and yields the path to the
    unpacked directory.

    The body of the with block is expected to inspect the wheel contents and
    possibly change it. If the body of the "with" block is successful, on
    exiting the with block the wheel contents are replaced with the updated
    contents of unpacked directory. If an exception is raised, then the original
    wheel is left unchanged.
    """
    with TemporaryDirectory() as temp_dir:
        unpack_wheel(wheel, Path(temp_dir), verbose=verbose)
        name, ver, _ = wheel.name.split("-", 2)
        wheel_dir_name = f"{name}-{ver}"
        wheel_dir = Path(temp_dir) / wheel_dir_name
        yield wheel_dir
        wheel.unlink()
        pack_wheel(wheel_dir, wheel.parent, verbose=verbose)


def retag_wheel(
    wheel_path: Path,
    platform: str,
    *,
    python: str | None = None,
    abi: str | None = None,
) -> Path:
    extra_flags = []
    if python:
        extra_flags += ["--python-tag", python]
    if abi:
        extra_flags += ["--abi-tag", abi]
    result = subprocess.run(
        [
            sys.executable,
            "-m",
            "wheel",
            "tags",
            wheel_path,
            "--platform-tag",
            platform,
            "--remove",
            *extra_flags,
        ],
        check=False,
        encoding="utf-8",
        capture_output=True,
    )
    if result.returncode != 0:
        logger.error("ERROR: Retagging wheel %s to %s failed", wheel_path, platform)
        exit_with_stdio(result)
    return wheel_path.parent / result.stdout.splitlines()[-1].strip()


def extract_wheel_metadata_file(wheel_path: Path, output_path: Path) -> None:
    """Extracts the METADATA file from the given wheel and writes it to the
    output path.

    Raises a RuntimeError if the METADATA file does not exist.

    For a wheel called "NAME-VERSION-...", the METADATA file is expected to be
    found in a directory inside the wheel archive, whose name starts with NAME
    and ends with ".dist-info". See:
    https://packaging.python.org/en/latest/specifications/binary-distribution-format/#file-contents
    """
    with ZipFile(wheel_path, mode="r") as wheel:
        pkg_name = wheel_path.name.split("-", 1)[0]
        dist_info_dir = get_wheel_dist_info_dir(wheel, pkg_name)
        metadata_path = f"{dist_info_dir}/METADATA"
        try:
            wheel.getinfo(metadata_path).filename = output_path.name
            wheel.extract(metadata_path, output_path.parent)
        except KeyError as err:
            raise RuntimeError(f"METADATA file not found for {pkg_name}") from err


def get_wheel_dist_info_dir(wheel: ZipFile, pkg_name: str) -> str:
    """Returns the path of the contained .dist-info directory.

    Raises a RuntimeError if the directory is not found, more than
    one is found, or it does not match the provided `pkg_name`.

    Adapted from:
    https://github.com/pypa/pip/blob/ea727e4d6ab598f34f97c50a22350febc1214a97/src/pip/_internal/utils/wheel.py#L38
    """

    # Zip file path separators must be /
    subdirs = {name.split("/", 1)[0] for name in wheel.namelist()}
    info_dirs = [subdir for subdir in subdirs if subdir.endswith(".dist-info")]

    if len(info_dirs) == 0:
        raise RuntimeError(f".dist-info directory not found for {pkg_name}")

    if len(info_dirs) > 1:
        raise RuntimeError(
            f"multiple .dist-info directories found for {pkg_name}: {', '.join(info_dirs)}"
        )

    (info_dir,) = info_dirs

    info_dir_name = canonicalize_package_name(info_dir)
    canonical_name = canonicalize_package_name(pkg_name)

    if not info_dir_name.startswith(canonical_name):
        raise RuntimeError(
            f".dist-info directory {info_dir!r} does not start with {canonical_name!r}"
        )

    return info_dir


def check_wasm_magic_number(file_path: Path) -> bool:
    WASM_BINARY_MAGIC = b"\0asm"
    with file_path.open(mode="rb") as file:
        return file.read(4) == WASM_BINARY_MAGIC


def search_pyproject_toml(
    curdir: str | Path, max_depth: int = 10
) -> tuple[Path, dict[str, Any]] | tuple[None, None]:
    """
    Recursively search for the pyproject.toml file in the parent directories.
    """

    # We want to include "curdir" in parent_dirs, so add a garbage suffix
    parent_dirs = (Path(curdir) / "garbage").parents[:max_depth]

    for base in parent_dirs:
        pyproject_file = base / "pyproject.toml"

        if not pyproject_file.is_file():
            continue

        try:
            with pyproject_file.open("rb") as f:
                configs = tomllib.load(f)
                return pyproject_file, configs
        except tomllib.TOMLDecodeError as e:
            raise ValueError(f"Could not parse {pyproject_file}.") from e

    return None, None


def to_bool(value: str) -> bool:
    """
    Convert a string to a boolean value. Useful for parsing environment variables.
    """
    return value.lower() not in {"", "0", "false", "no", "off"}


def download_and_unpack_archive(
    url: str, path: Path, descr: str, *, exists_ok: bool = False
) -> None:
    """
    Download the cross-build environment from the given URL and extract it to the given path.

    Parameters
    ----------
    url
        URL to download the cross-build environment from.
    path
        Path to extract the cross-build environment to.
        If the path already exists, raise an error.
    """
    logger.info("Downloading %s from %s", descr, url)

    if not exists_ok and path.exists():
        raise FileExistsError(f"Path {path} already exists")

    try:
        resp = urlopen(url)
        data = resp.read()
    except Exception as e:
        raise ValueError(f"Failed to download {descr} from {url}") from e

    # FIXME: requests makes a verbose output (see: https://github.com/pyodide/pyodide/issues/4810)
    # r = requests.get(url)

    # if r.status_code != 200:
    #     raise ValueError(
    #         f"Failed to download cross-build environment from {url} (status code: {r.status_code})"
    #     )

    with TemporaryDirectory() as temp_dir:
        f_path = Path(temp_dir) / "temp.tar"
        f_path.write_bytes(data)
        with warnings.catch_warnings():
            # Python 3.12-3.13 emits a DeprecationWarning when using shutil.unpack_archive without a filter,
            # but filter doesn't work well for zip files, so we suppress the warning until we find a better solution.
            # https://github.com/python/cpython/issues/112760
            warnings.simplefilter("ignore")
            shutil.unpack_archive(str(f_path), path)


def retrying_rmtree(d):
    """Sometimes rmtree fails with OSError: Directory not empty

    Try again a few times if this happens.
    See: https://github.com/python/cpython/issues/128076
    """
    for _ in range(3):
        try:
            return shutil.rmtree(d)
        except OSError as e:
            if e.strerror == "Directory not empty":
                # wait a bit and try again up to 3 tries
                time.sleep(0.01)
            else:
                raise
    raise RuntimeError(f"shutil.rmtree('{d}') failed with ENOTEMPTY three times")



import json
import os
import shutil
import subprocess as sp
import sys
import traceback
from collections.abc import Callable, Iterator, Mapping, Sequence
from contextlib import contextmanager
from itertools import chain
from pathlib import Path
from typing import Literal, cast

from build import BuildBackendException, ConfigSettingsType
from build.env import DefaultIsolatedEnv
from packaging.requirements import Requirement

from pyodide_build import _f2c_fixes, common, pywasmcross, uv_helper
from pyodide_build.build_env import (
    get_build_flag,
    get_host_build_flag,
    get_hostsitepackages,
    get_pyversion,
    get_unisolated_packages,
    platform,
)
from pyodide_build.spec import _BuildSpecExports
from pyodide_build.vendor._pypabuild import (
    _STYLES,
    _DefaultIsolatedEnv,
    _error,
    _handle_build_error,
    _ProjectBuilder,
)

# corresponding env variables for symlinks
SYMLINK_ENV_VARS = {
    "cc": "CC",
    "c++": "CXX",
    "ld": "LD",
    "lld": "LLD",
    "ar": "AR",
    "gcc": "GCC",
    "ranlib": "RANLIB",
    "strip": "STRIP",
    "gfortran": "FC",  # https://mesonbuild.com/Reference-tables.html#compiler-and-linker-selection-variables
    "cmake": "CMAKE_EXECUTABLE",  # For scikit-build to find cmake (https://github.com/scikit-build/scikit-build-core/pull/603)
}


def _gen_runner(
    cross_build_env: Mapping[str, str],
    isolated_build_env: _DefaultIsolatedEnv = None,
) -> Callable[[Sequence[str], str | None, Mapping[str, str] | None], None]:
    """
    This returns a slightly modified version of default subprocess runner that pypa/build uses.
    pypa/build prepends the virtual environment's bin directory to the PATH environment variable.
    This is problematic because it shadows the pywasmcross compiler wrappers for cmake, meson, etc.

    This function prepends the compiler wrapper directory to the PATH again so that our compiler wrappers
    are searched first.

    Parameters
    ----------
    cross_build_env
        The cross build environment for pywasmcross.
    isolated_build_env
        The isolated build environment created by pypa/build.
    """

    def _runner(cmd, cwd=None, extra_environ=None):
        env = os.environ.copy()
        if extra_environ:
            env.update(extra_environ)

        # Some build dependencies like cmake, meson installs binaries to this directory
        # and we should add it to the PATH so that they can be found.
        if isolated_build_env:
            env["BUILD_ENV_SCRIPTS_DIR"] = isolated_build_env.scripts_dir
        else:
            # For non-isolated builds, set a fallback path or use the current Python path
            import sysconfig

            scripts_dir = sysconfig.get_path("scripts")
            env["BUILD_ENV_SCRIPTS_DIR"] = scripts_dir

        env["PATH"] = f"{cross_build_env['COMPILER_WRAPPER_DIR']}:{env['PATH']}"
        # For debugging: Uncomment the following line to print the build command
        # print("Build backend call:", " ".join(str(x) for x in cmd), file=sys.stderr)
        sp.check_call(cmd, cwd=cwd, env=env)

    return _runner


def symlink_unisolated_packages(env: DefaultIsolatedEnv) -> None:
    from pyodide_build.build_env import get_build_flag, get_unisolated_packages

    pyversion = get_pyversion()
    site_packages_path = f"lib/{pyversion}/site-packages"
    env_site_packages = Path(env.path) / site_packages_path
    sysconfigdata_name = get_build_flag("SYSCONFIG_NAME")
    sysconfigdata_path = (
        Path(get_build_flag("TARGETINSTALLDIR"))
        / f"sysconfigdata/{sysconfigdata_name}.py"
    )

    env_site_packages.mkdir(parents=True, exist_ok=True)
    shutil.copy(sysconfigdata_path, env_site_packages)
    host_site_packages = Path(get_hostsitepackages())
    for name in get_unisolated_packages():
        for path in chain(
            host_site_packages.glob(f"{name}*"), host_site_packages.glob(f"_{name}*")
        ):
            (env_site_packages / path.name).unlink(missing_ok=True)
            (env_site_packages / path.name).symlink_to(path)


def remove_avoided_requirements(
    requires: set[str], avoided_requirements: set[str] | list[str]
) -> set[str]:
    for reqstr in list(requires):
        req = Requirement(reqstr)
        for avoid_name in set(avoided_requirements):
            if avoid_name == req.name.lower():
                requires.remove(reqstr)
    return requires


def install_reqs(
    build_env: Mapping[str, str], env: DefaultIsolatedEnv, reqs: set[str]
) -> None:
    IGNORED_BUILD_REQUIREMENTS = [
        pkg.strip() for pkg in get_host_build_flag("IGNORED_BUILD_REQUIREMENTS").split()
    ]
    # propagate PIP config from build_env to current environment
    with common.replace_env(
        os.environ | {k: v for k, v in build_env.items() if k.startswith("PIP")}
    ):
        env.install(
            remove_avoided_requirements(
                reqs,
                get_unisolated_packages() + IGNORED_BUILD_REQUIREMENTS,
            )
        )


def _build_in_isolated_env(
    build_env: Mapping[str, str],
    srcdir: Path,
    outdir: str,
    distribution: Literal["sdist", "wheel"],
    config_settings: ConfigSettingsType,
) -> str:
    # For debugging: The following line disables removal of the isolated venv.
    # It will be left in the /tmp folder and can be inspected or entered as
    # needed.
    # _DefaultIsolatedEnv.__exit__ = lambda self, *args: print("Skipping removing isolated env in", self.path)
    installer = "uv" if uv_helper.should_use_uv() else "pip"
    with _DefaultIsolatedEnv(installer=installer) as env:
        env = cast(_DefaultIsolatedEnv, env)
        builder = _ProjectBuilder.from_isolated_env(
            env,
            srcdir,
            runner=_gen_runner(build_env, env),
        )

        # first install the build dependencies
        symlink_unisolated_packages(env)
        install_reqs(build_env, env, builder.build_system_requires)
        build_reqs: set[str] | None = None
        try:
            build_reqs = builder.get_requires_for_build(
                distribution,
            )
        except BuildBackendException:
            pass

        if not build_reqs:
            # get_requires_for_build in native env failed. Maybe trying to
            # execute get_requires_for_build in the cross build environment will
            # work?

            # This case is used in pygame-ce. In native env, the setup.py picks
            # up native SDL2 config, then fails. In the cross env, it correctly
            # picks up Emscripten SDL2 config.
            # TODO: Add test coverage.
            with common.replace_env(build_env):
                build_reqs = builder.get_requires_for_build(
                    distribution,
                    config_settings,
                )

        install_reqs(build_env, env, build_reqs)

        with common.replace_env(build_env):
            return builder.build(
                distribution,
                outdir,
                config_settings,
            )


def _build_in_current_env(
    build_env: Mapping[str, str],
    srcdir: Path,
    outdir: str,
    distribution: Literal["sdist", "wheel"],
    config_settings: ConfigSettingsType,
    skip_dependency_check: bool = False,
) -> str:
    with common.replace_env(build_env):
        builder = _ProjectBuilder(srcdir, runner=_gen_runner(build_env))

        if not skip_dependency_check:
            missing = builder.check_dependencies(distribution, config_settings or {})
            if missing:
                dependencies = common._format_missing_dependencies(missing)
                _error(f"Missing dependencies: {dependencies}")

        return builder.build(
            distribution,
            outdir,
            config_settings,
        )


def parse_backend_flags(backend_flags: str | list[str]) -> ConfigSettingsType:
    config_settings: dict[str, str | list[str]] = {}

    if isinstance(backend_flags, str):
        backend_flags = backend_flags.split()

    for arg in backend_flags:
        setting, _, value = arg.partition("=")
        if setting not in config_settings:
            config_settings[setting] = value
            continue

        cur_value = config_settings[setting]
        if isinstance(cur_value, str):
            config_settings[setting] = [cur_value, value]
        else:
            cur_value.append(value)
    return config_settings


def make_command_wrapper_symlinks(symlink_dir: Path) -> dict[str, str]:
    """
    Create symlinks that make pywasmcross look like a compiler.

    Parameters
    ----------
    symlink_dir
        The directory where the symlinks will be created.

    Returns
    -------
    The dictionary of compiler environment variables that points to the symlinks.
    """
    # For maintainers:
    # - you can set "_f2c_fixes_wrapper" variable in pyproject.toml
    # in order to change the script to use when cross-compiling
    # this is only for maintainers and *should* not be used by others

    pywasmcross_exe = symlink_dir / "pywasmcross.py"
    pywasmcross_origin = pywasmcross.__file__
    shutil.copy2(pywasmcross_origin, pywasmcross_exe)
    pywasmcross_exe.chmod(0o755)

    f2c_fixes_exe = symlink_dir / "_f2c_fixes.py"
    f2c_fixes_origin = get_build_flag("_F2C_FIXES_WRAPPER") or _f2c_fixes.__file__
    shutil.copy2(f2c_fixes_origin, f2c_fixes_exe)

    env = {}
    for symlink in pywasmcross.SYMLINKS:
        symlink_path = symlink_dir / symlink
        if os.path.lexists(symlink_path) and not symlink_path.exists():
            # remove broken symlink so it can be re-created
            symlink_path.unlink()

        symlink_path.symlink_to(pywasmcross_exe)
        if symlink in SYMLINK_ENV_VARS:
            env[SYMLINK_ENV_VARS[symlink]] = str(symlink_path)

    return env


# TODO: a context manager is no longer needed here
@contextmanager
def _create_symlink_dir(
    env: dict[str, str],
    build_dir: Path,
):
    # Leave the symlinks in the build directory. This helps with reproducing.
    symlink_dir = build_dir / "pywasmcross_symlinks"
    shutil.rmtree(symlink_dir, ignore_errors=True)
    symlink_dir.mkdir()
    yield symlink_dir


@contextmanager
def get_build_env(
    env: dict[str, str],
    *,
    pkgname: str,
    cflags: str,
    cxxflags: str,
    ldflags: str,
    target_install_dir: str,
    exports: _BuildSpecExports,
    build_dir: Path | None = None,
    no_isolation: bool = False,
) -> Iterator[dict[str, str]]:
    """
    Returns a dict of environment variables that should be used when building
    a package with pypa/build.
    """
    from pyodide_build.build_env import get_build_flag

    kwargs = {
        "pkgname": pkgname,
        "cflags": cflags,
        "cxxflags": cxxflags,
        "ldflags": ldflags,
        "target_install_dir": target_install_dir,
    }

    args = common.environment_substitute_args(kwargs, env)
    args["exports"] = exports
    env = env.copy()

    with _create_symlink_dir(env, build_dir) as symlink_dir:
        env.update(make_command_wrapper_symlinks(symlink_dir))
        sysconfig_dir = Path(get_build_flag("TARGETINSTALLDIR")) / "sysconfigdata"
        args["PYTHONPATH"] = sys.path + [str(symlink_dir), str(sysconfig_dir)]
        args["orig__name__"] = __name__
        args["pythoninclude"] = get_build_flag("PYTHONINCLUDE")
        args["PATH"] = env["PATH"]
        args["abi"] = get_build_flag("PYODIDE_ABI_VERSION")

        pywasmcross_env = json.dumps(args)
        # Store into environment variable and to disk. In most cases we will
        # load from the environment variable but if some other tool filters
        # environment variables we will load from disk instead.
        env["PYWASMCROSS_ARGS"] = pywasmcross_env
        (symlink_dir / "pywasmcross_env.json").write_text(pywasmcross_env)

        env["_PYTHON_HOST_PLATFORM"] = platform()
        env["_PYTHON_SYSCONFIGDATA_NAME"] = get_build_flag("SYSCONFIG_NAME")
        env["PYTHONPATH"] = str(sysconfig_dir)
        env["COMPILER_WRAPPER_DIR"] = str(symlink_dir)

        yield env


def build(
    srcdir: Path,
    outdir: Path,
    build_env: Mapping[str, str],
    config_settings: ConfigSettingsType,
    isolation: bool = True,
    skip_dependency_check: bool = False,
) -> str:
    try:
        with _handle_build_error():
            if isolation:
                built = _build_in_isolated_env(
                    build_env, srcdir, str(outdir), "wheel", config_settings
                )
            else:
                built = _build_in_current_env(
                    build_env,
                    srcdir,
                    str(outdir),
                    "wheel",
                    config_settings,
                    skip_dependency_check,
                )
            print("{bold}{green}Successfully built {}{reset}".format(built, **_STYLES))
            return built
    except Exception as e:  # pragma: no cover
        tb = traceback.format_exc().strip("\n")
        print("\n{dim}{}{reset}\n".format(tb, **_STYLES))
        _error(str(e))
        sys.exit(1)



import re
from pathlib import Path

import typer

from pyodide_build._py_compile import _py_compile_archive, _py_compile_archive_dir


def main(
    path: Path = typer.Argument(
        ..., help="Path to the input wheel or a folder with wheels or zip files."
    ),
    silent: bool = typer.Option(False, help="Silent mode, do not print anything."),
    keep: bool = typer.Option(False, help="Keep the original wheel / zip file."),
    compression_level: int = typer.Option(
        6, help="Compression level to use for the created zip file"
    ),
    exclude: str = typer.Option(
        "",
        help="List of files to exclude from compilation, works only for directories. Defaults to no files.",
    ),
) -> None:
    """Compile .py files to .pyc in a wheel, a zip file, or a folder with wheels or zip files.

    If the provided folder contains the `pyodide-lock.json` file, it will be
    rewritten with the updated wheel / zip file paths and sha256 checksums.
    """
    if not path.exists():
        typer.echo(f"Error: {path} does not exist")
        raise typer.Exit(1)

    # Convert the comma / space separated strings to lists
    excludes = [
        item.strip() for item in re.split(r",|\s", exclude) if item.strip() != ""
    ]

    if path.is_file():
        if path.suffix not in [".whl", ".zip"]:
            typer.echo(
                f"Error: only .whl and .zip files are supported, got {path.name}"
            )
            raise typer.Exit(1)

        _py_compile_archive(
            path, verbose=not silent, keep=keep, compression_level=compression_level
        )
    elif path.is_dir():
        _py_compile_archive_dir(
            path,
            verbose=not silent,
            keep=keep,
            compression_level=compression_level,
            excludes=excludes,
        )
    else:
        typer.echo(f"{path=} is not a file or a directory")



import json
import os
import subprocess
import sys
import textwrap
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path
from types import TracebackType
from typing import Any, TextIO

from pyodide_build.build_env import (
    get_build_environment_vars,
    get_pyodide_root,
)
from pyodide_build.common import exit_with_stdio
from pyodide_build.logger import logger


class BashRunnerWithSharedEnvironment:
    """Run multiple bash scripts with persistent environment.

    Environment is stored to "env" member between runs. This can be updated
    directly to adjust the environment, or read to get variables.
    """

    def __init__(self, env: dict[str, str] | None = None) -> None:
        if env is None:
            env = dict(os.environ)

        self._reader: TextIO | None
        self._fd_write: int | None
        self.env: dict[str, str] = env

    def __enter__(self) -> "BashRunnerWithSharedEnvironment":
        fd_read, self._fd_write = os.pipe()
        self._reader = os.fdopen(fd_read, "r")
        return self

    def run_unchecked(self, cmd: str, **opts: Any) -> subprocess.CompletedProcess[str]:
        assert self._fd_write is not None
        assert self._reader is not None

        write_env_pycode = ";".join(
            [
                "import os",
                "import json",
                f'os.write({self._fd_write}, json.dumps(dict(os.environ)).encode() + b"\\n")',
            ]
        )
        write_env_shell_cmd = f"{sys.executable} -c '{write_env_pycode}'"
        full_cmd = f"{cmd}\n{write_env_shell_cmd}"
        result = subprocess.run(
            ["bash", "-ce", full_cmd],
            check=False,
            pass_fds=[self._fd_write],
            env=self.env,
            encoding="utf8",
            **opts,
        )
        if result.returncode == 0:
            self.env = json.loads(self._reader.readline())
        return result

    def run(
        self,
        cmd: str | None,
        *,
        script_name: str,
        cwd: Path | str | None = None,
        **opts: Any,
    ) -> subprocess.CompletedProcess[str] | None:
        """Run a bash script. Any keyword arguments are passed on to subprocess.run."""
        if not cmd:
            return None
        if cwd is None:
            cwd = Path.cwd()
        cwd = Path(cwd).absolute()
        logger.info("Running %s in %s", script_name, str(cwd))
        opts["cwd"] = cwd
        result = self.run_unchecked(cmd, **opts)
        if result.returncode != 0:
            logger.error("ERROR: %s failed", script_name)
            logger.error(textwrap.indent(cmd, "    "))
            exit_with_stdio(result)
        return result

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Free the file descriptors."""

        if self._fd_write:
            os.close(self._fd_write)
            self._fd_write = None
        if self._reader:
            self._reader.close()
            self._reader = None


@contextmanager
def get_bash_runner(
    extra_envs: dict[str, str],
) -> Iterator[BashRunnerWithSharedEnvironment]:
    pyodide_root = get_pyodide_root()
    env = get_build_environment_vars(pyodide_root)
    env.update(extra_envs)

    with BashRunnerWithSharedEnvironment(env=env) as b:
        # Working in-tree, add emscripten toolchain into PATH and set ccache
        if Path(pyodide_root, "pyodide_env.sh").exists():
            b.run(
                f"source {pyodide_root}/pyodide_env.sh",
                script_name="source pyodide_env",
                stderr=subprocess.DEVNULL,
            )

        yield b



import json
import shutil
import subprocess
import sys
from pathlib import Path

from pyodide_lock import PyodideLockSpec

from pyodide_build import build_env, uv_helper
from pyodide_build.common import download_and_unpack_archive
from pyodide_build.create_package_index import create_package_index
from pyodide_build.logger import logger
from pyodide_build.xbuildenv_releases import (
    CrossBuildEnvReleaseSpec,
    cross_build_env_metadata_url,
    load_cross_build_env_metadata,
)

CDN_BASE = "https://cdn.jsdelivr.net/pyodide/v{version}/full/"
PYTHON_VERSION_MARKER_FILE = ".build-python-version"


class CrossBuildEnvManager:
    """
    Manager for the cross-build environment.
    """

    def __init__(self, env_dir: str | Path, metadata_url: str | None = None) -> None:
        """
        Parameters
        ----------
        env_dir
            The directory to store the cross-build environments.
        metadata_url
            URL to the metadata file that contains the information about the available
            cross-build environments. If not specified, the default metadata file is used.
        """
        self.env_dir = Path(env_dir).resolve()
        self.metadata_url = metadata_url or cross_build_env_metadata_url()

        try:
            self.env_dir.mkdir(parents=True, exist_ok=True)
        except Exception as e:
            raise ValueError(
                f"Failed to create cross-build environment at {self.env_dir}"
            ) from e

    @property
    def symlink_dir(self):
        """
        Returns the path to the symlink that points to the currently used xbuildenv version.
        """
        return self.env_dir / "xbuildenv"

    @property
    def pyodide_root(self) -> Path:
        """
        Returns the path to the pyodide-root directory inside the xbuildenv directory.
        """
        return self.symlink_dir.resolve() / "xbuildenv" / "pyodide-root"

    @property
    def current_version(self) -> str | None:
        """
        Returns the currently used xbuildenv version.
        """
        if not self.symlink_dir.exists():
            return None

        return self.symlink_dir.resolve().name

    def _find_remote_release(self, version: str) -> CrossBuildEnvReleaseSpec:
        """
        Find the cross-build environment release with the given version from remote metadata.
        """
        metadata = load_cross_build_env_metadata(self.metadata_url)
        return metadata.get_release(version)

    def _path_for_version(self, version: str) -> Path:
        """Returns the path to the xbuildenv for the given version."""
        return self.env_dir / version

    def list_versions(self) -> list[str]:
        """
        List the downloaded xbuildenv versions.

        TODO: add a parameter to list only compatible versions
        """
        versions = []
        for version_dir in self.env_dir.glob("*"):
            if not version_dir.is_dir() or version_dir == self.symlink_dir:
                continue

            versions.append(version_dir.name)

        return sorted(versions)

    def use_version(self, version: str) -> None:
        """
        Select the xbuildenv version to use.

        This creates a symlink to the selected version in the xbuildenv directory.

        Parameters
        ----------
        version
            The version of xbuildenv to use.
        """
        logger.info("Using Pyodide cross-build environment version: %s", version)

        version_path = self._path_for_version(version)
        if not version_path.exists():
            raise ValueError(
                f"Cannot find cross-build environment version {version}, available versions: {self.list_versions()}"
            )

        symlink_dir = self.symlink_dir

        if symlink_dir.exists():
            if symlink_dir.is_symlink():
                # symlink to a directory, expected case
                symlink_dir.unlink()
            elif symlink_dir.is_dir():
                # real directory, for backwards compatibility
                shutil.rmtree(symlink_dir)
            else:
                # file. This should not happen unless the user manually created a file
                # but we will remove it anyway
                symlink_dir.unlink()

        symlink_dir.symlink_to(version_path)

    def install(
        self,
        version: str | None = None,
        *,
        url: str | None = None,
        skip_install_cross_build_packages: bool = False,
        force_install: bool = False,
    ) -> Path:
        """
        Install cross-build environment.

        Parameters
        ----------
        version
            The version of the cross-build environment to install. If not specified,
            use the same version as the current version of pyodide-build.
        url
            URL to download the cross-build environment from.
            The URL should point to a tarball containing the cross-build environment.
            This is useful for testing unreleased version of the cross-build environment.

            Warning: if you are downloading from a version that is not the same
            as the current version of pyodide-build, make sure that the cross-build
            environment is compatible with the current version of Pyodide.
        skip_install_cross_build_packages
            If True, skip installing the cross-build packages. This is mostly for testing purposes.
        force_install
            If True, force the installation even if the cross-build environment is not compatible

        Returns
        -------
        Path to the root directory for the cross-build environment.
        """

        if url and version:
            raise ValueError("Cannot specify both version and url")

        if url:
            version = _url_to_version(url)
            download_url = url
        # if default version is specified in the configuration, use that
        elif not version and (default_url := self._get_default_xbuildenv_url()):
            version = _url_to_version(default_url)
            download_url = default_url
        else:
            version = version or self._find_latest_version()

            local_versions = build_env.local_versions()
            release = self._find_remote_release(version)
            if not force_install and not release.is_compatible(
                python_version=local_versions["python"],
                pyodide_build_version=local_versions["pyodide-build"],
            ):
                raise ValueError(
                    f"Version {version} is not compatible with the current environment."
                )

            download_url = release.url

        download_path = self._path_for_version(version)

        if download_path.exists():
            logger.info(
                "The cross-build environment already exists at '%s', skipping download",
                download_path,
            )
        else:
            download_and_unpack_archive(
                download_url, download_path, "Pyodide cross-build environment"
            )

        try:
            # there is an redundant directory "xbuildenv" inside the xbuildenv archive
            # TODO: remove the redundant directory from the archive
            xbuildenv_root = download_path / "xbuildenv"
            xbuildenv_pyodide_root = xbuildenv_root / "pyodide-root"
            install_marker = download_path / ".installed"
            if not install_marker.exists():
                logger.info(
                    "Installing Pyodide cross-build environment to %s", download_path
                )

                if not skip_install_cross_build_packages:
                    self._install_cross_build_packages(
                        xbuildenv_root, xbuildenv_pyodide_root
                    )

                if not url:
                    # If installed from url, skip creating the PyPI index (version is not known)
                    self._create_package_index(xbuildenv_pyodide_root, version)

            install_marker.touch()
            self.use_version(version)
            self._add_version_marker()
        except Exception as e:
            # if the installation failed, remove the downloaded directory
            shutil.rmtree(download_path)
            raise e

        return xbuildenv_pyodide_root

    def _find_latest_version(self) -> str:
        """
        Find the latest compatible cross-build environment release.
        """
        metadata = load_cross_build_env_metadata(self.metadata_url)
        local = build_env.local_versions()
        latest = metadata.get_latest_compatible_release(
            python_version=local["python"],
            pyodide_build_version=local["pyodide-build"],
        )

        if not latest:
            # Check for Python version mismatch
            python_versions = [
                v.python_version_tuple[:2] for v in metadata.list_compatible_releases()
            ]
            pyver = tuple(int(x) for x in local["python"].split("."))
            if pyver > python_versions[0]:
                latest_supported = ".".join(str(x) for x in python_versions[0])
                raise ValueError(
                    f"Python version {local['python']} is not yet supported. The newest supported version of Python is {latest_supported}."
                )

            if pyver < python_versions[-1]:
                oldest_supported = ".".join(str(x) for x in python_versions[-1])
                raise ValueError(
                    f"Python version {local['python']} is too old. The oldest supported version of Python is {oldest_supported}."
                )

            raise ValueError(
                f"Python version {local['python']} is not compatible with pyodide build version {local['pyodide-build']}"
            )

        return latest.version

    def _get_default_xbuildenv_url(self) -> str:
        """
        Get the default URL for the cross-build environment. If not specified, return empty string (no default).
        """
        return build_env.get_host_build_flag("DEFAULT_CROSS_BUILD_ENV_URL")

    def _install_cross_build_packages(
        self, xbuildenv_root: Path, xbuildenv_pyodide_root: Path
    ) -> None:
        """
        Install package that are used in the cross-build environment.

        Parameters
        ----------
        xbuildenv_root
            Path to the xbuildenv directory.
        xbuildenv_pyodide_root
            Path to the pyodide-root directory inside the xbuildenv directory.
        """
        host_site_packages = self._host_site_packages_dir(xbuildenv_pyodide_root)
        host_site_packages.mkdir(exist_ok=True, parents=True)

        install_prefix = (
            [
                uv_helper.find_uv_bin(),
                "pip",
                "install",
            ]
            if uv_helper.should_use_uv()
            else [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--no-user",
            ]
        )

        result = subprocess.run(
            [
                *install_prefix,
                "-r",
                str(xbuildenv_root / "requirements.txt"),
                "--target",
                str(host_site_packages),
            ],
            capture_output=True,
            encoding="utf8",
            check=False,
        )

        if result.returncode != 0:
            raise RuntimeError(
                f"Failed to install cross-build packages: {result.stderr}"
            )

        # Copy the site-packages-extras (coming from the cross-build-files meta.yaml
        # key) over the site-packages directory with the newly installed packages.
        shutil.copytree(
            xbuildenv_root / "site-packages-extras",
            host_site_packages,
            dirs_exist_ok=True,
        )

    def _host_site_packages_dir(
        self, xbuildenv_pyodide_root: Path | None = None
    ) -> Path:
        """
        Returns the path to the hostsitepackages directory in the xbuild environment.
        This is inferred using the current version of the xbuild environment,
        but can optionally be overridden by passing the pyodide root directory as a parameter.

        Parameters
        ----------
        xbuildenv_pyodide_root
            The path to the pyodide root directory inside the xbuild environment.
        """

        if xbuildenv_pyodide_root is None:
            xbuildenv_pyodide_root = self.pyodide_root

        return Path(
            build_env.get_build_environment_vars(pyodide_root=xbuildenv_pyodide_root)[
                "HOSTSITEPACKAGES"
            ]
        )

    def _create_package_index(self, xbuildenv_pyodide_root: Path, version: str) -> None:
        """
        Create the PyPI index for the packages in the xbuild environment.
        TODO: Creating the PyPI Index is not required for the xbuild environment to work, so maybe we can
              move this to a separate command (to pyodide venv?)
        """

        cdn_base = CDN_BASE.format(version=version)
        lockfile_path = xbuildenv_pyodide_root / "dist" / "pyodide-lock.json"

        if not lockfile_path.exists():
            logger.warning(
                "Pyodide lockfile not found at %s. Skipping PyPI index creation",
                lockfile_path,
            )
            return

        lockfile = PyodideLockSpec(**json.loads(lockfile_path.read_bytes()))
        create_package_index(lockfile.packages, xbuildenv_pyodide_root, cdn_base)

    def uninstall_version(self, version: str | None) -> str:
        """
        Uninstall the installed xbuildenv version.

        Parameters
        ----------
        version
            The version of xbuildenv to uninstall.
        """
        if version is None:
            version = self.current_version

        if version is None:
            raise ValueError("No xbuildenv version is currently in use")

        version_path = self._path_for_version(version)

        # if the target version is the current version, remove the symlink
        # to prevent symlinking to a non-existent directory
        if self.symlink_dir.resolve() == version_path:
            self.symlink_dir.unlink()

        if version_path.is_dir():
            shutil.rmtree(version_path)
        else:
            raise ValueError(
                f"Cannot find cross-build environment version {version}, available versions: {self.list_versions()}"
            )

        return version

    def _add_version_marker(self) -> None:
        """
        Store the Python version in the xbuildenv directory, so we can check compatibility later.
        """
        if not self.symlink_dir.is_dir():
            raise ValueError("cross-build env directory does not exist")

        version_file = self.symlink_dir / PYTHON_VERSION_MARKER_FILE
        version_file.write_text(build_env.local_versions()["python"])

    def check_version_marker(self):
        if not self.symlink_dir.is_dir():
            raise ValueError("cross-build env directory does not exist")

        version_file = self.symlink_dir / PYTHON_VERSION_MARKER_FILE
        if not version_file.exists():
            raise ValueError("Python version marker file not found")

        version_local = build_env.local_versions()["python"]
        version_on_install = version_file.read_text().strip()
        if version_on_install != version_local:
            raise ValueError(
                f"local Python version ({version_local}) does not match the Python version ({version_on_install}) "
                "used to create the Pyodide cross-build environment. "
                "Please switch back to the original Python version, "
                "or reinstall the xbuildenv, by running `pyodide xbuildenv uninstall` and then `pyodide xbuildenv install`"
            )


def _url_to_version(url: str) -> str:
    return url.replace("://", "_").replace(".", "_").replace("/", "_")
